from __future__ import annotations

from copy import deepcopy
from typing import List, Tuple

from .puzzle import Puzzle
from .rle import Rle
from .utilities import is_blank


class PuzzleResizer:
    def add_row_top(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        old_body = parsed_board
        old_height = height

        new_height = height + 1
        new_width = width
        new_body = new_height * new_width * [Puzzle.VISIBLE_FLOOR]

        for x in range(0, new_width):
            for y in range(0, old_height):
                new_body[index_1d(x, y + 1, new_width)] = old_body[
                    index_1d(x, y, new_width)
                ]

        return new_body, new_width, new_height

    def add_row_bottom(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        old_body = parsed_board
        old_height = height

        new_height = height + 1
        new_width = width
        new_body = new_height * new_width * [Puzzle.VISIBLE_FLOOR]

        for x in range(0, new_width):
            for y in range(0, old_height):
                new_body[index_1d(x, y, new_width)] = old_body[
                    index_1d(x, y, new_width)
                ]

        return new_body, new_width, new_height

    def add_column_left(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        old_body = parsed_board
        old_width = width

        new_height = height
        new_width = width + 1
        new_body = new_height * new_width * [Puzzle.VISIBLE_FLOOR]

        for x in range(0, old_width):
            for y in range(0, new_height):
                new_body[index_1d(x + 1, y, new_width)] = old_body[
                    index_1d(x, y, old_width)
                ]

        return new_body, new_width, new_height

    def add_column_right(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        old_body = parsed_board
        old_width = width

        new_height = height
        new_width = width + 1
        new_body = new_height * new_width * [Puzzle.VISIBLE_FLOOR]

        for x in range(0, old_width):
            for y in range(0, new_height):
                new_body[index_1d(x, y, new_width)] = old_body[
                    index_1d(x, y, old_width)
                ]

        return new_body, new_width, new_height

    def remove_row_top(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        old_body = parsed_board

        new_height = height - 1
        new_width = width
        if new_height <= 0:
            return [], 0, 0
        new_body = new_height * new_width * [Puzzle.VISIBLE_FLOOR]

        for x in range(0, new_width):
            for y in range(0, new_height):
                new_body[index_1d(x, y, new_width)] = old_body[
                    index_1d(x, y + 1, new_width)
                ]

        return new_body, new_width, new_height

    def remove_row_bottom(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        old_body = parsed_board

        new_height = height - 1
        new_width = width
        if new_height <= 0:
            return [], 0, 0
        new_body = new_height * new_width * [Puzzle.VISIBLE_FLOOR]

        for x in range(0, new_width):
            for y in range(0, new_height):
                new_body[index_1d(x, y, new_width)] = old_body[
                    index_1d(x, y, new_width)
                ]

        return new_body, new_width, new_height

    def remove_column_left(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        old_body = parsed_board
        old_width = width

        new_height = height
        new_width = width - 1
        if new_width <= 0:
            return [], 0, 0
        new_body = new_height * new_width * [Puzzle.VISIBLE_FLOOR]

        for x in range(0, new_width):
            for y in range(0, new_height):
                new_body[index_1d(x, y, new_width)] = old_body[
                    index_1d(x + 1, y, old_width)
                ]

        return new_body, new_width, new_height

    def remove_column_right(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        old_body = parsed_board
        old_width = width

        new_height = height
        new_width = width - 1
        if new_width <= 0:
            return [], 0, 0
        new_body = new_height * new_width * [Puzzle.VISIBLE_FLOOR]

        for x in range(0, new_width):
            for y in range(0, new_height):
                new_body[index_1d(x, y, new_width)] = old_body[
                    index_1d(x, y, old_width)
                ]

        return new_body, new_width, new_height

    def trim_left(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        amount = width
        for y in range(0, height):
            border_found = False
            for x in range(0, width):
                border_found = Puzzle.is_border_element(
                    parsed_board[index_1d(x, y, width)]
                )
                if border_found and x < amount:
                    amount = x
                    break

        new_body = deepcopy(parsed_board)
        new_width = width
        new_height = height
        for _ in range(0, amount):
            new_body, new_width, new_height = self.remove_column_left(
                new_body, new_width, new_height
            )

        return new_body, new_width, new_height

    def trim_right(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        new_body = deepcopy(parsed_board)
        new_width = width
        new_height = height

        new_body, new_width, new_height = self.reverse_columns(
            new_body, new_width, new_height
        )
        new_body, new_width, new_height = self.trim_left(
            new_body, new_width, new_height
        )
        new_body, new_width, new_height = self.reverse_columns(
            new_body, new_width, new_height
        )

        return new_body, new_width, new_height

    def trim_top(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        amount = height
        for x in range(0, width):
            border_found = False
            for y in range(0, height):
                border_found = Puzzle.is_border_element(
                    parsed_board[index_1d(x, y, width)]
                )
                if border_found:
                    if y < amount:
                        amount = y
                    break

        new_body = deepcopy(parsed_board)
        new_width = width
        new_height = height

        for _ in range(0, amount):
            new_body, new_width, new_height = self.remove_row_top(
                new_body, new_width, new_height
            )

        return new_body, new_width, new_height

    def trim_bottom(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        new_body = deepcopy(parsed_board)
        new_width = width
        new_height = height

        new_body, new_width, new_height = self.reverse_rows(
            new_body, new_width, new_height
        )
        new_body, new_width, new_height = self.trim_top(new_body, new_width, new_height)
        new_body, new_width, new_height = self.reverse_rows(
            new_body, new_width, new_height
        )

        return new_body, new_width, new_height

    def reverse_rows(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        old_body = deepcopy(parsed_board)

        new_body = width * height * [Puzzle.VISIBLE_FLOOR]

        for x in range(0, width):
            for y in range(0, height):
                new_body[index_1d(x, y, width)] = str(
                    old_body[index_1d(x, height - y - 1, width)]
                )

        return new_body, width, height

    def reverse_columns(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        from ..game import index_1d

        old_body = deepcopy(parsed_board)

        new_body = width * height * [Puzzle.VISIBLE_FLOOR]

        for x in range(0, width):
            for y in range(0, height):
                new_body[index_1d(x, y, width)] = str(
                    old_body[index_1d(width - x - 1, y, width)]
                )

        return new_body, width, height


class PuzzleParser:
    def parse(self, board: str) -> List[str]:
        return self.cleaned_board_lines(board)

    @classmethod
    def calculate_width(cls, strings: List[str]) -> int:
        """Width of list of strings as length of longest string in that list."""
        width = 0
        for line in strings:
            if len(line) > width:
                width = len(line)
        return width

    @classmethod
    def normalize_width(cls, strings: List[str], fill_chr: str = " ") -> List[str]:
        """
        Normalizes length of strings in ``string_list``.

        All strings are modified to be as long as the longest one in list. Missing
        characters in string are appended using ``fill_chr``.
        """
        width = cls.calculate_width(strings)
        return [l + (fill_chr * (width - len(l))) for l in strings]

    @classmethod
    def cleaned_board_lines(cls, line: str) -> List[str]:
        """
        Converts line into width-normalized, decoded list of board lines suitable for
        conversion into proper board object.
        """
        if is_blank(line):
            return []

        if not Puzzle.is_board(line):
            raise ValueError("Illegal characters found in board string")

        return cls.normalize_width(Rle.decode(line).split("\n"))


class PuzzlePrinter:
    def print(
        self,
        parsed_board: List[str],
        width: int,
        height: int,
        use_visible_floor: bool,
        rle_encode: bool,
    ) -> str:
        from ..game import index_1d

        retv_list: List[str] = []

        for y in range(height):
            tmp: str = ""
            for x in range(width):
                c = parsed_board[index_1d(x, y, width)]
                if Puzzle.is_empty_floor(c):
                    c = Puzzle.VISIBLE_FLOOR if use_visible_floor else Puzzle.FLOOR
                tmp += c

            retv_list.append(tmp)

        retv = "\n".join(retv_list)

        if rle_encode:
            retv = Rle.encode(retv)

        return retv



from sokoenginepy.game import BaseTessellation, Direction, PusherStep, Tessellation

from .autogenerated_tessellation import OctobanTessellationAutogeneratedSpecMixin
from .tessellation_spec_mixin import TessellationSpecMixin


class DescribeOctobanTessellation(
    TessellationSpecMixin, OctobanTessellationAutogeneratedSpecMixin
):

    illegal_directions = []

    legal_directions = [
        Direction.LEFT,
        Direction.RIGHT,
        Direction.UP,
        Direction.DOWN,
        Direction.NORTH_WEST,
        Direction.NORTH_EAST,
        Direction.SOUTH_WEST,
        Direction.SOUTH_EAST,
    ]

    legal_moves = [
        PusherStep(Direction.LEFT, False),
        PusherStep(Direction.RIGHT, False),
        PusherStep(Direction.UP, False),
        PusherStep(Direction.DOWN, False),
        PusherStep(Direction.NORTH_WEST, False),
        PusherStep(Direction.NORTH_EAST, False),
        PusherStep(Direction.SOUTH_WEST, False),
        PusherStep(Direction.SOUTH_EAST, False),
        PusherStep(Direction.LEFT, True),
        PusherStep(Direction.RIGHT, True),
        PusherStep(Direction.UP, True),
        PusherStep(Direction.DOWN, True),
        PusherStep(Direction.NORTH_WEST, True),
        PusherStep(Direction.NORTH_EAST, True),
        PusherStep(Direction.SOUTH_WEST, True),
        PusherStep(Direction.SOUTH_EAST, True),
    ]

    legal_characters = [
        "l",
        "r",
        "u",
        "d",
        "w",
        "n",
        "s",
        "e",
        "L",
        "R",
        "U",
        "D",
        "W",
        "N",
        "S",
        "E",
    ]

    tessellation = BaseTessellation.instance(Tessellation.OCTOBAN)



import pytest

from sokoenginepy.game import PusherStep


class TessellationSpecMixin:
    def test_converts_legal_pusher_steps_to_characters(self):
        for index, pusher_step in enumerate(self.legal_moves):
            assert (
                self.tessellation.pusher_step_to_char(pusher_step)
                == self.legal_characters[index]
            )

    def test_converts_legal_characters_to_pusher_steps(self):
        for index, pusher_step_str in enumerate(self.legal_characters):
            assert (
                self.tessellation.char_to_pusher_step(pusher_step_str)
                == self.legal_moves[index]
            )

    def test_it_raises_on_illegal_character(self):
        with pytest.raises(ValueError):
            self.tessellation.char_to_pusher_step("z")

    def test_it_raises_when_converting_illegal_direction(self):
        for index, illegal_direction in enumerate(self.illegal_directions):
            move = PusherStep(illegal_direction, False)
            with pytest.raises(ValueError):
                self.tessellation.pusher_step_to_char(move)
            push = PusherStep(illegal_direction, True)
            with pytest.raises(ValueError):
                self.tessellation.pusher_step_to_char(push)



#!/usr/bin/env python

# Requires: 'pyexcel-ods3 >= 0.2.0',

import os
import textwrap
from collections import OrderedDict
from pathlib import Path
from string import Template

from pyexcel_ods3 import get_data

# Directory where this file is found
_SELF_DIR = Path(__file__).parent.absolute().resolve()

SOURCE_ROOT = (_SELF_DIR / ".." / ".." / "..").resolve()
TESTS_DIR = SOURCE_ROOT / "tests"
RES_DIR = TESTS_DIR / "fixtures"
TEST_CASES_DIR = RES_DIR / "test_cases"
BOARD_TESTS_DIR = TESTS_DIR / "io"
TESSELLATION_TESTS_DIR = TESTS_DIR / "game"


class SpecGenerator:
    VARIANTS = ["sokoban", "trioban", "hexoban", "octoban"]
    TEST_TYPES = ["tessellation"]

    DIRECTIONS = ["l", "r", "u", "d", "nw", "sw", "ne", "se"]
    DIRECTIONS_HASH = dict(
        zip(
            DIRECTIONS,
            [
                "Direction.LEFT",
                "Direction.RIGHT",
                "Direction.UP",
                "Direction.DOWN",
                "Direction.NORTH_WEST",
                "Direction.SOUTH_WEST",
                "Direction.NORTH_EAST",
                "Direction.SOUTH_EAST",
            ],
        )
    )
    RESULTS_HASH = dict(zip(DIRECTIONS, ["result_{0}".format(d) for d in DIRECTIONS]))

    def generate(self):
        self.write_header()

        for variant in type(self).VARIANTS:
            print("Generating test cases for {0}...".format(variant))
            test_cases = self.get_test_cases(variant)

            for test_type in self.TEST_TYPES:
                self.output = []
                self.generate_test_class(variant, test_type)
                for test_case in test_cases:
                    test_case["test_type"] = test_type
                    self.generate_test_case_output(test_case)

                if test_type == "tessellation":
                    with open(str(self.tessellation_output_file_path), "a") as f:
                        for line in self.output:
                            f.write(line)
                        f.write("\n")
                else:
                    raise RuntimeError(f"Unknown TEST_TYPES {test_type}")

    def write_header(self):
        s = """
            import pytest

            from sokoenginepy.game import (
                BaseTessellation,
                Direction,
                OctobanTessellation,
                Tessellation,
                TriobanTessellation,
                index_1d,
                is_on_board_1d,
            )
            from sokoenginepy.io import CellOrientation


            def triangle_points_down(position, board_width, board_height):
                return (
                    TriobanTessellation().cell_orientation(position, board_width, board_height)
                    == CellOrientation.TRIANGLE_DOWN
                )


            def is_octagon(position, board_width, board_height):
                return (
                    OctobanTessellation().cell_orientation(position, board_width, board_height)
                    == CellOrientation.OCTAGON
                )
        """

        with open(str(self.tessellation_output_file_path), "w") as f:
            f.write(textwrap.dedent(s).strip() + "\n\n\n")

    def get_test_cases(self, variant):
        data = get_data(str(self.input_file_path(variant)))
        sheet1 = data[list(data.keys())[0]]
        rows = [row[:12] for row in sheet1 if len(row) > 0 and len(row[0]) > 0]
        header = rows[0]
        raw_cases = [OrderedDict(zip(header, row)) for row in rows[1:]]

        cases = []
        for case in raw_cases:
            case["variant"] = variant
            case["width"] = int(case["board_dimensions"].split(",")[0].strip())
            case["height"] = int(case["board_dimensions"].split(",")[1].strip())
            case["row"] = int(case["test_position"].split(",")[1].strip())
            case["column"] = int(case["test_position"].split(",")[0].strip())
            cases.append(case)

        return cases

    def generate_test_class(self, variant, test_type):
        s = """
            class {0}{1}AutogeneratedSpecMixin:
                class Describe_neighbor_position:
        """.format(
            variant.capitalize(), test_type.capitalize()
        )
        self.output.append(textwrap.dedent(s).strip() + "\n")

    def generate_test_case_output(self, test_case):
        s = Template(
            """
            def test_autogenerated_$test_name(self):
                width = $width
                height = $height
                row = $row
                column = $column
                index = index_1d(column, row, width)
        """
        ).substitute(
            test_name=test_case["test_name"],
            width=test_case["width"],
            height=test_case["height"],
            row=test_case["row"],
            column=test_case["column"],
        )
        self.output.append(
            textwrap.indent(textwrap.dedent(s).strip() + "\n", prefix="        ")
        )

        if test_case["test_type"] == "tessellation":
            self.output.append(
                f'            t = BaseTessellation.instance(Tessellation.{test_case["variant"].upper()})'
            )
        else:
            raise RuntimeError(f'Unknown TEST_TYPES {test_case["test_type"]}')
        self.output.append("\n")

        if test_case["test_position_requrement"] != "nil":
            self.output.append("\n")
            self.output.append(
                "            assert {0}".format(
                    test_case["test_position_requrement"]
                    .replace("&&", "and")
                    .replace("||", "or")
                    .replace("!", "not ")
                )
            )
            self.output.append("\n\n")
        else:
            self.output.append("\n")

        for direction in self.DIRECTIONS:
            if test_case["test_type"] == "tessellation":
                if self.is_result_illegal_position(test_case, direction):
                    s = """
                        assert not is_on_board_1d(t.neighbor_position(index, {0}, width, height), width, height)
                    """.format(
                        self.DIRECTIONS_HASH[direction]
                    )

                elif self.is_result_illegal_direction(test_case, direction):
                    s = """
                        with pytest.raises(ValueError):
                            t.neighbor_position(index, {0}, width, height)
                    """.format(
                        self.DIRECTIONS_HASH[direction]
                    )
                else:
                    s = """
                        assert t.neighbor_position(index, {0}, width, height) == index_1d({1}, width)
                    """.format(
                        self.DIRECTIONS_HASH[direction],
                        test_case[self.RESULTS_HASH[direction]],
                    )
            else:
                if self.is_result_illegal_position(
                    test_case, direction
                ) or self.is_result_illegal_direction(test_case, direction):
                    s = """
                        assert not is_on_board_1d(b.neighbor(index, {0}), width, height)
                    """.format(
                        self.DIRECTIONS_HASH[direction]
                    )
                else:
                    s = """
                        assert b.neighbor(index, {0}) == index_1d({1}, width)
                    """.format(
                        self.DIRECTIONS_HASH[direction],
                        test_case[self.RESULTS_HASH[direction]],
                    )

            self.output.append(
                textwrap.indent(
                    textwrap.dedent(s).strip() + "\n", prefix="            "
                )
            )

        self.output.append("\n")

    def input_file_path(self, variant):
        return os.path.join(
            TEST_CASES_DIR, "{0}_test_cases.ods".format(variant.strip().lower())
        )

    @property
    def board_output_file_path(self):
        return os.path.join(BOARD_TESTS_DIR, "autogenerated_puzzle.py")

    @property
    def tessellation_output_file_path(self):
        return os.path.join(TESSELLATION_TESTS_DIR, "autogenerated_tessellation.py")

    def is_result_illegal_direction(self, test_case, direction):
        return test_case[self.RESULTS_HASH[direction]] == "IllegalDirection"

    def is_result_illegal_position(self, test_case, direction):
        return test_case[self.RESULTS_HASH[direction]] == "OfBoardPosition"


def main():
    generator = SpecGenerator()
    generator.generate()


if __name__ == "__main__":
    main()



from sokoenginepy.game import (
    Config,
    Direction,
    JumpCommand,
    MoveCommand,
    Mover,
    PusherStep,
    SelectPusherCommand,
    SolvingMode,
)

from .mover_spec import (
    forward_board,
    jump_dest,
    jumps,
    pusher_selections,
    reverse_board,
    undone_jumps,
    undone_pusher_selections,
)


class DescribeSelectPusherCommand:
    def it_executes_pusher_selection(
        self, forward_board, pusher_selections, undone_pusher_selections
    ):
        forward_mover = Mover(forward_board)
        command = SelectPusherCommand(forward_mover, Config.DEFAULT_PIECE_ID + 1)

        assert command.old_pusher_id == forward_mover.selected_pusher
        assert command.new_pusher_id == Config.DEFAULT_PIECE_ID + 1

        command.redo()
        assert command.moves in pusher_selections
        assert command.rendered in pusher_selections
        assert forward_mover.selected_pusher == Config.DEFAULT_PIECE_ID + 1

        command.undo()
        assert command.moves in pusher_selections
        assert command.rendered in undone_pusher_selections
        assert forward_mover.selected_pusher == Config.DEFAULT_PIECE_ID

        command.redo()
        assert command.moves in pusher_selections
        assert command.rendered in pusher_selections
        assert forward_mover.selected_pusher == Config.DEFAULT_PIECE_ID + 1

        command.undo()
        assert command.moves in pusher_selections
        assert command.rendered in undone_pusher_selections
        assert forward_mover.selected_pusher == Config.DEFAULT_PIECE_ID


class DescribeJumpCommand:
    def it_executes_jump(self, forward_board, jumps, undone_jumps, jump_dest):
        reverse_mover = Mover(forward_board, SolvingMode.REVERSE)
        command = JumpCommand(reverse_mover, jump_dest)

        assert command.initial_position == reverse_mover.board_manager.pusher_position(
            Config.DEFAULT_PIECE_ID
        )
        assert command.final_position == jump_dest

        command.redo()
        assert command.moves in jumps
        assert command.rendered in jumps
        assert (
            reverse_mover.board_manager.pusher_position(Config.DEFAULT_PIECE_ID) == jump_dest
        )

        command.undo()
        assert command.moves in jumps
        assert command.rendered in undone_jumps
        assert (
            reverse_mover.board_manager.pusher_position(Config.DEFAULT_PIECE_ID)
            == command.initial_position
        )

        command.redo()
        assert command.moves in jumps
        assert command.rendered in jumps
        assert (
            reverse_mover.board_manager.pusher_position(Config.DEFAULT_PIECE_ID) == jump_dest
        )

        command.undo()
        assert command.moves in jumps
        assert command.rendered in undone_jumps
        assert (
            reverse_mover.board_manager.pusher_position(Config.DEFAULT_PIECE_ID)
            == command.initial_position
        )


class DescribeMoveCommand:
    def it_executes_movement(self, forward_board):
        forward_mover = Mover(forward_board)
        command = MoveCommand(forward_mover, Direction.UP)

        command.redo()
        assert command.moves == [PusherStep(Direction.UP)]
        assert command.rendered == [PusherStep(Direction.UP)]

        command.undo()
        assert command.moves == [PusherStep(Direction.UP)]
        assert command.rendered == [PusherStep(Direction.DOWN)]

        command.redo()
        assert command.moves == [PusherStep(Direction.UP)]
        assert command.rendered == [PusherStep(Direction.UP)]

        command.undo()
        assert command.moves == [PusherStep(Direction.UP)]
        assert command.rendered == [PusherStep(Direction.DOWN)]



from __future__ import annotations

from collections import deque
from typing import Callable, Dict, List, Optional, Sequence, Set, Tuple, Union

import networkx as nx

from ..io import Puzzle
from .base_tessellation import BaseTessellation, Tessellation
from .board_cell import BoardCell
from .config import Config, Direction, GraphType

# (1, 0, {'direction': Direction.LEFT})
Edge = Tuple[int, int, Dict[str, Union[Direction, int]]]
BoardCellOrStr = Union[BoardCell, str]
Positions = List[int]
Directions = List[Direction]


class BoardGraph:
    """
    Board graph.

    Depending on how ``sokoenginepy`` was installed, it is using either ``NetworkX`` or
    ``Boost.Graph`` under the hood.
    """

    _KEY_CELL = "cell"
    _KEY_DIRECTION = "direction"
    _MAX_EDGE_WEIGHT = 100  # must be > len(Direction)

    def __init__(self, puzzle: Puzzle):
        if puzzle.width > Config.MAX_HEIGHT or puzzle.height > Config.MAX_WIDTH:
            raise ValueError(
                "Board width and height must be >= 0 and <= MAX_BOARD_WIDTH, "
                "MAX_BOARD_HEIGHT`!"
            )

        self._board_width = puzzle.width
        self._board_height = puzzle.height
        self._tessellation = puzzle.tessellation

        tessellation = BaseTessellation.instance(self._tessellation)
        if tessellation.graph_type == GraphType.DIRECTED:
            self._graph = nx.DiGraph()
        elif tessellation.graph_type == GraphType.DIRECTED_MULTI:
            self._graph = nx.MultiDiGraph()
        else:
            raise ValueError(f"Unknown graph_type: {tessellation.graph_type.name}!")

        for position in range(0, self.board_width * self.board_height):
            self._graph.add_node(
                position, **{self._KEY_CELL: BoardCell(puzzle[position])}
            )
        self.reconfigure_edges()

    def __getitem__(self, position: int) -> BoardCell:
        try:
            return self._graph.nodes[position][self._KEY_CELL]

        except KeyError as e:
            if isinstance(position, int) and position >= 0:
                raise IndexError from e
            else:
                raise

    def __setitem__(self, position: int, board_cell: BoardCellOrStr):
        try:
            if isinstance(board_cell, BoardCell):
                self._graph.nodes[position][self._KEY_CELL] = board_cell
            else:
                self._graph.nodes[position][self._KEY_CELL] = BoardCell(board_cell)

        except KeyError as e:
            if isinstance(position, int) and position >= 0:
                raise IndexError from e
            else:
                raise

    def __contains__(self, position: int) -> bool:
        return position in self._graph

    @property
    def tessellation(self) -> Tessellation:
        return self._tessellation

    def to_board_str(self, use_visible_floor=False, rle_encode=False) -> str:
        puzzle = Puzzle.instance_from(
            self._tessellation, self._board_width, self._board_height
        )

        for pos in range(self.vertices_count):
            puzzle[pos] = self[pos].to_str()

        return puzzle.to_board_str(use_visible_floor, rle_encode)

    def __str__(self) -> str:
        return self.to_board_str(True)

    @property
    def vertices_count(self) -> int:
        return self._graph.number_of_nodes()

    @property
    def size(self) -> int:
        return self._board_width * self._board_height

    @property
    def edges_count(self) -> int:
        return self._graph.number_of_edges()

    @property
    def board_width(self) -> int:
        return self._board_width

    @property
    def board_height(self) -> int:
        return self._board_height

    def has_edge(self, src: int, dst: int, direction: Direction) -> bool:
        retv = False
        out_edge: Edge
        try:
            if src is not None:
                for out_edge in self._graph.edges(src, data=True):
                    retv = retv or (
                        out_edge[1] == dst
                        and out_edge[2][self._KEY_DIRECTION] == direction
                    )
                    if retv:
                        break

        except (KeyError, IndexError, nx.NetworkXError):
            pass

        return retv

    def out_edges_count(self, src: int, dst: int) -> int:
        """
        Number of out-edges from ``src`` to ``dst``.

        Returns:
            Zero when no out edges exist or or any of positions is illegal type or out
            of bound board index.
        """
        try:
            # retv = self._graph.number_of_edges(src, dst)
            retv = len(self._graph[src][dst])
        except (KeyError, IndexError, nx.NetworkXError):
            retv = 0

        return retv

    def remove_all_edges(self):
        self._graph.remove_edges_from(list(self._graph.edges.keys()))

    def add_edge(self, src: int, neighbor: int, direction: Direction):
        """
        Adds edges between to existing positions.

        Raises:
            IndexError: ``src`` or ``neighbor`` off board
            KeyError: ``src`` or ``neighbor`` illegal values
        """
        if self[src] and self[neighbor]:
            self._graph.add_edge(src, neighbor, direction=direction)

    def out_edge_weight(self, target_position: int) -> int:
        """
        Calculates edge weight based on BoardCell on ``target_position``.

        Raises:
            IndexError: ``target_position`` off board
            KeyError: ``target_position`` illegal values
        """
        target_cell = self[target_position]

        weight = 1
        if target_cell and (
            target_cell.is_wall or target_cell.has_box or target_cell.has_pusher
        ):
            weight = self._MAX_EDGE_WEIGHT

        return weight

    def neighbor(self, src: int, direction: Direction) -> Optional[int]:
        """
        Calculates neighbor position in ``direction``.

        Returns:
            Target position or `None`.

        Raises:
            IndexError: ``src`` off board
            KeyError: ``src`` illegal values
        """
        out_edge: Edge
        if self[src]:
            for out_edge in self._graph.edges(src, data=True):
                if out_edge[2][self._KEY_DIRECTION] == direction:
                    return out_edge[1]

        return None

    def wall_neighbors(self, src: int) -> Positions:
        """
        Gets a list of all neighboring walls.

        Raises:
            IndexError: ``src`` off board
            KeyError: ``src`` illegal values
        """
        if self[src]:
            return [n for n in self._graph.neighbors(src) if self[n].is_wall]

        return []

    def all_neighbors(self, src: int) -> Positions:
        """
        Gets a list of all neighbors.

        Raises:
            IndexError: ``src`` off board
            KeyError: ``src`` illegal values
        """

        if self[src]:
            return list(self._graph.neighbors(src))

        return []

    def shortest_path(self, src: int, dst: int) -> Positions:
        """
        Calculates shortest path between two positions with all positions having equal
        weight.

        Raises:
            IndexError: ``src`` or ``dst`` off board
            KeyError: ``src`` or ``dst`` illegal values
        """

        if self[src] and self[dst]:
            edge: Edge
            for edge in self._graph.edges(data=True):
                edge[2]["weight"] = 1

            try:
                return nx.shortest_path(self._graph, src, dst, 1)
            except nx.NetworkXNoPath:
                return []

        return []

    def dijkstra_path(self, src: int, dst: int) -> Positions:
        """
        Calculates shortest path between two positions not passing through board
        obstacles (walls, other pushers, etc...).

        Raises:
            IndexError: ``src`` or ``dst`` off board
            KeyError: ``src`` or ``dst`` illegal values
        """
        if self[src] and self[dst]:
            edge: Edge
            for edge in self._graph.edges(data=True):
                edge[2]["weight"] = self.out_edge_weight(edge[1])

            try:
                return nx.dijkstra_path(self._graph, src, dst)
            except nx.NetworkXNoPath:
                return []

        return []

    def find_jump_path(self, src: int, dst: int) -> Positions:
        """
        Returns:
            List of positions through which pusher must pass when jumping

        Raises:
            IndexError: ``src`` or ``dst`` off board
            KeyError: ``src`` or ``dst`` illegal values
        """
        return self.shortest_path(src, dst)

    def find_move_path(self, src: int, dst: int) -> Positions:
        """
        Returns:
            List of positions through which pusher must pass when moving without
            pushing boxes

        Raises:
            IndexError: ``src`` or ``dst`` off board
            KeyError: ``src`` or ``dst`` illegal values
        """
        path = self.dijkstra_path(src, dst)

        retv = path[:1]
        for position in path[1:]:
            if self[position].can_put_pusher_or_box:
                retv.append(position)
            else:
                break
        if retv != path:
            return []
        return path

    def positions_path_to_directions_path(self, positions: Positions) -> Directions:
        """
        Converts path expressed as positions to one expressed as :class:`.Direction`.

        Raises:
            IndexError: Any of positions in ``positions`` off board
            KeyError: Any of positions in ``positions`` are illegal values
        """

        if positions:
            self[positions[0]]
        else:
            return []

        retv = []

        src_position_index = 0
        out_edge: Edge
        for dst in positions[1:]:
            src_position = positions[src_position_index]
            src_position_index += 1

            if self[src_position] and self[dst]:
                for out_edge in self._graph.edges(src_position, data=True):
                    if out_edge[1] == dst:
                        retv.append(out_edge[2][self._KEY_DIRECTION])

        return retv

    def mark_play_area(self):
        """
        Sets flag on all :class:`.BoardCell` in graph that are playable: reachable by
        any box or any pusher.
        """
        piece_positions = []
        for position in range(0, self.vertices_count):
            if self[position].has_box or self[position].has_pusher:
                self[position].is_in_playable_area = True
                piece_positions.append(position)
            else:
                self[position].is_in_playable_area = False

        for piece_position in piece_positions:
            reachables = self._reachables(
                root=piece_position, is_obstacle_cb=lambda x: self[x].is_wall
            )

            for reachable_position in reachables:
                self[reachable_position].is_in_playable_area = True

    def positions_reachable_by_pusher(
        self, pusher_position: int, excluded_positions: Optional[Positions] = None
    ) -> Positions:
        """
        Returns:
            List of positions that are reachable by pusher standing on
            ``pusher_position``.
        """
        return self._reachables(
            root=pusher_position,
            is_obstacle_cb=lambda x: not self[x].can_put_pusher_or_box,
            excluded_positions=excluded_positions,
        )

    def normalized_pusher_position(
        self, pusher_position: int, excluded_positions: Optional[Positions] = None
    ) -> int:
        """
        Returns:
            Top-left position reachable by pusher.
        """
        reachables = self.positions_reachable_by_pusher(
            pusher_position=pusher_position, excluded_positions=excluded_positions
        )
        if reachables:
            return min(reachables)
        else:
            return pusher_position

    def path_destination(self, src: int, directions: Directions) -> int:
        if not directions:
            self[src]

        retv = src
        for direction in directions:
            next_target = self.neighbor(retv, direction)
            if next_target:
                retv = next_target
            else:
                break
        return retv

    def reconfigure_edges(self):
        tessellation = BaseTessellation.instance(self.tessellation)

        self.remove_all_edges()
        for src in range(self.vertices_count):
            for direction in tessellation.legal_directions:
                neighbor_position = tessellation.neighbor_position(
                    src, direction, self.board_width, self.board_height
                )
                if neighbor_position is not None:
                    self.add_edge(src, neighbor_position, direction)

    _CurrentReachables = Sequence[int]
    _ToInspectVertices = Sequence[int]
    _ExcludedVertices = Optional[List[int]]
    _AnimationFrameCallback = Callable[
        [int, _CurrentReachables, _ToInspectVertices, _ExcludedVertices], None
    ]

    def _reachables(
        self,
        root: int,
        excluded_positions: Optional[Positions] = None,
        is_obstacle_cb: Optional[Callable[[int], bool]] = None,
        add_animation_frame_cb: Optional[_AnimationFrameCallback] = None,
    ) -> List[int]:
        """
        Calculates all positions reachable from ``root``.

        Args:
            root: initial position for search
            excluded_positions: these positions will be marked as unreachable without
                calculating their status
            is_obstacle_cb: callable that checks if given position on graph is obstacle
            add_animation_frame_cb: if not None, this callable will be called after
                each step of search. Useful for visualization of algorithm and debugging
        """
        excluded: Set[int] = set(excluded_positions) if excluded_positions else set()

        if is_obstacle_cb is None:
            is_obstacle_cb = lambda x: not self[x].can_put_pusher_or_box

        reachables = deque()

        if self[root]:
            visited = self.vertices_count * [False]
            visited[root] = True
            to_inspect = deque([root])

            while len(to_inspect) > 0:
                current_position = to_inspect.popleft()

                if current_position == root or current_position not in excluded:
                    reachables.append(current_position)

                for neighbor in self.all_neighbors(current_position):
                    if not visited[neighbor]:
                        if not is_obstacle_cb(neighbor):
                            to_inspect.append(neighbor)
                        visited[neighbor] = True

                if add_animation_frame_cb is not None:
                    add_animation_frame_cb(
                        current_position, reachables, to_inspect, excluded_positions
                    )

        if root in excluded:
            return [pos for pos in reachables if pos != root]
        else:
            return list(reachables)



import pytest

from sokoenginepy.game import Config, Direction, PusherStep

from ..fixtures import PusherStepFactory


@pytest.fixture
def pusher_step():
    return PusherStepFactory(direction=Direction.LEFT, box_moved=False)


class DescribePusherStep:
    class Describe_init:
        def it_sets_all_attributes(self):
            pusher_step = PusherStepFactory(direction=Direction.RIGHT, box_moved=True)

            assert pusher_step.direction == Direction.RIGHT
            assert not pusher_step.is_move
            assert not pusher_step.is_pusher_selection
            assert not pusher_step.is_jump
            assert pusher_step.pusher_id == Config.DEFAULT_PIECE_ID
            assert pusher_step.moved_box_id is None

        def it_validates_parameters(self):
            with pytest.raises(ValueError):
                PusherStep(box_moved=True, is_pusher_selection=True, is_jump=True)
            with pytest.raises(ValueError):
                PusherStep(moved_box_id=42, is_pusher_selection=True, is_jump=True)

            with pytest.raises(ValueError):
                PusherStep(box_moved=True, is_jump=True)
            with pytest.raises(ValueError):
                PusherStep(moved_box_id=42, is_jump=True)

            with pytest.raises(ValueError):
                PusherStep(box_moved=True, is_pusher_selection=True)
            with pytest.raises(ValueError):
                PusherStep(moved_box_id=42, is_pusher_selection=True)

            with pytest.raises(ValueError):
                PusherStep(is_jump=True, is_pusher_selection=True)

    class Describe_moved_box_id:
        def test_get_returns_none_if_move_is_not_push_or_pull(self, pusher_step):
            pusher_step.moved_box_id = Config.DEFAULT_PIECE_ID
            pusher_step.is_push_or_pull = False
            assert pusher_step.moved_box_id is None

        def test_get_returns_box_that_moved_if_move_is_push_or_pull(self, pusher_step):
            pusher_step.is_push_or_pull = True
            pusher_step.moved_box_id = Config.DEFAULT_PIECE_ID + 42
            assert pusher_step.moved_box_id == Config.DEFAULT_PIECE_ID + 42

        def test_when_setting_to_none_it_also_resets_push_or_pull_flag(
            self, pusher_step
        ):
            pusher_step.is_push_or_pull = True
            pusher_step.moved_box_id = None

            assert pusher_step.moved_box_id is None
            assert not pusher_step.is_push_or_pull
            assert pusher_step.is_move

        def test_when_setting_to_not_none_it_also_sets_push_or_pull_flag(
            self, pusher_step
        ):
            pusher_step.is_push_or_pull = False
            pusher_step.moved_box_id = Config.DEFAULT_PIECE_ID + 42

            assert pusher_step.moved_box_id == Config.DEFAULT_PIECE_ID + 42
            assert pusher_step.is_push_or_pull
            assert not pusher_step.is_move

        def it_sets_it_to_none_if_illegal_value_provided(self, pusher_step):
            pusher_step.moved_box_id = 4
            assert pusher_step.moved_box_id == 4
            pusher_step.moved_box_id = -42
            assert pusher_step.moved_box_id is None

            pusher_step.moved_box_id = 4
            assert pusher_step.moved_box_id == 4
            pusher_step.moved_box_id = None
            assert pusher_step.moved_box_id is None

            pusher_step.moved_box_id = 4
            assert pusher_step.moved_box_id == 4
            pusher_step.moved_box_id = "ZOMG!"
            assert pusher_step.moved_box_id is None

            pusher_step.moved_box_id = 4
            assert pusher_step.moved_box_id == 4
            pusher_step.moved_box_id = 0
            assert pusher_step.moved_box_id is None

            m = PusherStep(moved_box_id=-42)
            assert m.moved_box_id is None

            m = PusherStep(moved_box_id=None)
            assert m.moved_box_id is None

            m = PusherStep(moved_box_id="ZOMG")
            assert m.moved_box_id is None

            m = PusherStep(moved_box_id=0)
            assert m.moved_box_id is None

    class Describe_pusher_id:
        def it_returns_id_of_pusher_that_performed_movement(self, pusher_step):
            pusher_step.pusher_id = Config.DEFAULT_PIECE_ID + 42
            assert pusher_step.pusher_id == Config.DEFAULT_PIECE_ID + 42

        def it_sets_it_to_default_if_illegal_value_provided(self, pusher_step):
            pusher_step.pusher_id = 4
            assert pusher_step.pusher_id == 4
            pusher_step.pusher_id = -42
            assert pusher_step.pusher_id == Config.DEFAULT_PIECE_ID

            pusher_step.pusher_id = 4
            assert pusher_step.pusher_id == 4
            pusher_step.pusher_id = None
            assert pusher_step.pusher_id == Config.DEFAULT_PIECE_ID

            pusher_step.pusher_id = 4
            assert pusher_step.pusher_id == 4
            pusher_step.pusher_id = "ZOMG!"
            assert pusher_step.pusher_id == Config.DEFAULT_PIECE_ID

    class Describe_is_move:
        def it_returns_true_if_box_was_not_moved(self, pusher_step):
            pusher_step.is_move = True
            assert pusher_step.is_move

        def it_returns_false_if_box_was_moved(self, pusher_step):
            pusher_step.is_move = True
            pusher_step.is_push_or_pull = True
            assert not pusher_step.is_move

        def it_returns_false_if_pusher_jumped(self, pusher_step):
            pusher_step.is_move = True
            pusher_step.is_jump = True
            assert not pusher_step.is_move

        def it_returns_false_if_pusher_was_selected(self, pusher_step):
            pusher_step.is_move = True
            pusher_step.is_pusher_selection = True
            assert not pusher_step.is_move

        def test_setter_sets_to_state_of_box_not_moved(self, pusher_step):
            pusher_step.is_move = True
            assert pusher_step.is_move

        def test_setter_sets_moved_box_id_to_none_when_setting_to_true(
            self, pusher_step
        ):
            pusher_step.is_move = True
            assert pusher_step.moved_box_id is None

        def test_setter_resets_other_flags_when_setting_to_true(self, pusher_step):
            pusher_step.is_push_or_pull = True
            pusher_step.is_move = True
            assert not pusher_step.is_push_or_pull

            pusher_step.is_pusher_selection = True
            pusher_step.is_move = True
            assert not pusher_step.is_pusher_selection

            pusher_step.is_jump = True
            pusher_step.is_move = True
            assert not pusher_step.is_jump

        def test_setter_assumes_push_when_setting_to_false(self, pusher_step):
            pusher_step.is_move = False
            assert pusher_step.is_push_or_pull
            assert not pusher_step.is_pusher_selection
            assert not pusher_step.is_jump

    class Describe_is_push_or_pull:
        def it_returns_true_if_box_was_moved(self, pusher_step):
            pusher_step.is_push_or_pull = True
            assert pusher_step.is_push_or_pull

        def it_returns_false_if_box_was_not_moved(self, pusher_step):
            pusher_step.is_push_or_pull = True
            pusher_step.is_move = True
            assert not pusher_step.is_push_or_pull

        def it_returns_false_if_pusher_jumped(self, pusher_step):
            pusher_step.is_push_or_pull = True
            pusher_step.is_jump = True
            assert not pusher_step.is_push_or_pull

        def it_returns_false_if_pusher_was_selected(self, pusher_step):
            pusher_step.is_push_or_pull = True
            pusher_step.is_pusher_selection = True
            assert not pusher_step.is_push_or_pull

        def test_setter_sets_to_state_of_box_moved(self, pusher_step):
            pusher_step.is_push_or_pull = True
            assert pusher_step.is_push_or_pull

        def test_setter_sets_moved_box_id_to_none_when_setting_to_false(
            self, pusher_step
        ):
            pusher_step.is_push_or_pull = False
            assert pusher_step.moved_box_id is None

        def test_setter_resets_other_flags_when_setting_to_true(self, pusher_step):
            pusher_step.is_move = True
            pusher_step.is_push_or_pull = True
            assert not pusher_step.is_move

            pusher_step.is_pusher_selection = True
            pusher_step.is_push_or_pull = True
            assert not pusher_step.is_pusher_selection

            pusher_step.is_jump = True
            pusher_step.is_push_or_pull = True
            assert not pusher_step.is_jump

        def test_setter_doesnt_touch_other_flags_when_setting_to_false(
            self, pusher_step
        ):
            pusher_step.is_move = True
            pusher_step.is_push_or_pull = False
            assert pusher_step.is_move

            pusher_step.is_pusher_selection = True
            pusher_step.is_push_or_pull = False
            assert pusher_step.is_pusher_selection

            pusher_step.is_jump = True
            pusher_step.is_push_or_pull = False
            assert pusher_step.is_jump

    class Describe_is_pusher_selection:
        def it_returns_true_if_pusher_was_selected(self, pusher_step):
            pusher_step.is_pusher_selection = True
            assert pusher_step.is_pusher_selection

        def it_returns_false_if_pusher_was_not_selected(self, pusher_step):
            pusher_step.is_pusher_selection = True
            pusher_step.is_move = True
            assert not pusher_step.is_pusher_selection

        def it_returns_false_if_pusher_jumped(self, pusher_step):
            pusher_step.is_pusher_selection = True
            pusher_step.is_jump = True
            assert not pusher_step.is_pusher_selection

        def it_returns_false_if_pusher_was_selected(self, pusher_step):
            pusher_step.is_pusher_selection = True
            pusher_step.is_push_or_pull = True
            assert not pusher_step.is_pusher_selection

        def test_setter_sets_to_state_of_pusher_was_selected(self, pusher_step):
            pusher_step.is_pusher_selection = True
            assert pusher_step.is_pusher_selection

        def test_setter_sets_moved_box_id_to_none_when_setting_to_true(
            self, pusher_step
        ):
            pusher_step.is_pusher_selection = True
            assert pusher_step.moved_box_id is None

        def test_setter_resets_other_flags_when_setting_to_true(self, pusher_step):
            pusher_step.is_move = True
            pusher_step.is_pusher_selection = True
            assert not pusher_step.is_move

            pusher_step.is_push_or_pull = True
            pusher_step.is_pusher_selection = True
            assert not pusher_step.is_push_or_pull

            pusher_step.is_jump = True
            pusher_step.is_pusher_selection = True
            assert not pusher_step.is_jump

        def test_setter_doesnt_touch_other_flags_when_setting_to_false(
            self, pusher_step
        ):
            pusher_step.is_move = True
            pusher_step.is_pusher_selection = False
            assert pusher_step.is_move

            pusher_step.is_push_or_pull = True
            pusher_step.is_pusher_selection = False
            assert pusher_step.is_push_or_pull

            pusher_step.is_jump = True
            pusher_step.is_pusher_selection = False
            assert pusher_step.is_jump

    class Describe_is_jump:
        def it_returns_true_if_pusher_jumped(self, pusher_step):
            pusher_step.is_jump = True
            assert pusher_step.is_jump

        def it_returns_false_if_pusher_moved_insted_of_jumping(self, pusher_step):
            pusher_step.is_jump = True
            pusher_step.is_move = True
            assert not pusher_step.is_jump

        def it_returns_false_if_pusher_was_selected(self, pusher_step):
            pusher_step.is_jump = True
            pusher_step.is_pusher_selection = True
            assert not pusher_step.is_jump

        def it_returns_false_if_box_was_moved(self, pusher_step):
            pusher_step.is_jump = True
            pusher_step.is_push_or_pull = True
            assert not pusher_step.is_jump

        def test_setter_sets_to_state_of_pusher_jumped(self, pusher_step):
            pusher_step.is_jump = True
            assert pusher_step.is_jump

        def test_setter_sets_moved_box_id_to_none_when_setting_to_true(
            self, pusher_step
        ):
            pusher_step.is_jump = True
            assert pusher_step.moved_box_id is None

        def test_setter_resets_other_flags_when_setting_to_true(self, pusher_step):
            pusher_step.is_move = True
            pusher_step.is_jump = True
            assert not pusher_step.is_move

            pusher_step.is_push_or_pull = True
            pusher_step.is_jump = True
            assert not pusher_step.is_push_or_pull

            pusher_step.is_pusher_selection = True
            pusher_step.is_jump = True
            assert not pusher_step.is_pusher_selection

        def test_setter_doesnt_touch_other_flags_when_setting_to_false(
            self, pusher_step
        ):
            pusher_step.is_move = True
            pusher_step.is_jump = False
            assert pusher_step.is_move

            pusher_step.is_push_or_pull = True
            pusher_step.is_jump = False
            assert pusher_step.is_push_or_pull

            pusher_step.is_pusher_selection = True
            pusher_step.is_jump = False
            assert pusher_step.is_pusher_selection



import factory

from sokoenginepy.game import PusherStep, Direction

from .misc import fake


class PusherStepFactory(factory.Factory):
    class Meta:
        model = PusherStep

    box_moved = factory.LazyAttribute(lambda x: fake.boolean())
    direction = factory.LazyAttribute(lambda x: fake.random_element(list(Direction)))



from sokoenginepy.game import Direction


class DescribeDirection_opposite:
    def it_returns_opposite_direction(self):
        assert Direction.UP.opposite == Direction.DOWN
        assert Direction.DOWN.opposite == Direction.UP
        assert Direction.LEFT.opposite == Direction.RIGHT
        assert Direction.RIGHT.opposite == Direction.LEFT
        assert Direction.NORTH_WEST.opposite == Direction.SOUTH_EAST
        assert Direction.NORTH_EAST.opposite == Direction.SOUTH_WEST
        assert Direction.SOUTH_WEST.opposite == Direction.NORTH_EAST
        assert Direction.SOUTH_EAST.opposite == Direction.NORTH_WEST



from sokoenginepy.game import BaseTessellation, Direction, PusherStep, Tessellation

from .autogenerated_tessellation import SokobanTessellationAutogeneratedSpecMixin
from .tessellation_spec_mixin import TessellationSpecMixin


class DescribeSokobanTessellation(
    TessellationSpecMixin, SokobanTessellationAutogeneratedSpecMixin
):

    illegal_directions = [
        Direction.NORTH_WEST,
        Direction.NORTH_EAST,
        Direction.SOUTH_WEST,
        Direction.SOUTH_EAST,
    ]

    legal_directions = [Direction.LEFT, Direction.RIGHT, Direction.UP, Direction.DOWN]

    legal_moves = [
        PusherStep(Direction.LEFT, False),
        PusherStep(Direction.RIGHT, False),
        PusherStep(Direction.UP, False),
        PusherStep(Direction.DOWN, False),
        PusherStep(Direction.LEFT, True),
        PusherStep(Direction.RIGHT, True),
        PusherStep(Direction.UP, True),
        PusherStep(Direction.DOWN, True),
    ]

    legal_characters = ["l", "r", "u", "d", "L", "R", "U", "D"]

    tessellation = BaseTessellation.instance(Tessellation.SOKOBAN)



import textwrap
from itertools import permutations

import pytest

from sokoenginepy.game import (
    BoardGraph,
    BoardManager,
    BoardState,
    CellAlreadyOccupiedError,
    Config,
    SokobanPlus,
    index_1d,
)
from sokoenginepy.io import SokobanPuzzle


@pytest.fixture
def puzzle():
    #   0123456789012345678
    data = """
        ----#####----------
        ----#--@#----------
        ----#$--#----------
        --###--$##---------
        --#--$-$-#---------
        ###-#-##-#---######
        #---#-##-#####--..#
        #-$--$----------..#
        #####-###-#@##--..#
        ----#-----#########
        ----#######--------
    """
    data = textwrap.dedent(data.lstrip("\n").rstrip())
    return SokobanPuzzle(board=data)


@pytest.fixture
def board_graph(puzzle):
    return BoardGraph(puzzle)


@pytest.fixture
def switched_puzzle():
    #   0123456789012345678
    data = """
        ----#####----------
        ----#--@#----------
        ----#.--#----------
        --###--.##---------
        --#--.-.-#---------
        ###-#-##-#---######
        #---#-##-#####--$$#
        #-.--.----------$$#
        #####-###-#@##--$$#
        ----#-----#########
        ----#######--------
    """
    data = textwrap.dedent(data.lstrip("\n").rstrip())
    return SokobanPuzzle(board=data)


@pytest.fixture
def pushers_positions(puzzle):
    return {
        Config.DEFAULT_PIECE_ID: index_1d(7, 1, puzzle.width),
        Config.DEFAULT_PIECE_ID + 1: index_1d(11, 8, puzzle.width),
    }


@pytest.fixture
def invalid_pusher_position():
    return index_1d(11, 8, 42)


@pytest.fixture
def boxes_positions(puzzle):
    return {
        Config.DEFAULT_PIECE_ID: index_1d(5, 2, puzzle.width),
        Config.DEFAULT_PIECE_ID + 1: index_1d(7, 3, puzzle.width),
        Config.DEFAULT_PIECE_ID + 2: index_1d(5, 4, puzzle.width),
        Config.DEFAULT_PIECE_ID + 3: index_1d(7, 4, puzzle.width),
        Config.DEFAULT_PIECE_ID + 4: index_1d(2, 7, puzzle.width),
        Config.DEFAULT_PIECE_ID + 5: index_1d(5, 7, puzzle.width),
    }


@pytest.fixture
def invalid_box_position():
    return index_1d(5, 7, 42)


@pytest.fixture
def goals_positions(puzzle):
    return {
        Config.DEFAULT_PIECE_ID: index_1d(16, 6, puzzle.width),
        Config.DEFAULT_PIECE_ID + 1: index_1d(17, 6, puzzle.width),
        Config.DEFAULT_PIECE_ID + 2: index_1d(16, 7, puzzle.width),
        Config.DEFAULT_PIECE_ID + 3: index_1d(17, 7, puzzle.width),
        Config.DEFAULT_PIECE_ID + 4: index_1d(16, 8, puzzle.width),
        Config.DEFAULT_PIECE_ID + 5: index_1d(17, 8, puzzle.width),
    }


@pytest.fixture
def invalid_goal_position():
    return index_1d(17, 8, 42)


@pytest.fixture
def all_solutions(goals_positions):
    def calc():
        for boxes_positions in permutations(goals_positions.values()):
            yield BoardState(
                boxes_positions=list(boxes_positions), pushers_positions=[]
            )

    return list(calc())


@pytest.fixture
def sokoban_plus_solutions():
    return [
        BoardState(
            boxes_positions=[149, 130, 131, 150, 168, 169], pushers_positions=[]
        ),
        BoardState(
            boxes_positions=[149, 130, 131, 150, 169, 168], pushers_positions=[]
        ),
        BoardState(
            boxes_positions=[149, 130, 131, 168, 150, 169], pushers_positions=[]
        ),
        BoardState(
            boxes_positions=[149, 130, 131, 168, 169, 150], pushers_positions=[]
        ),
        BoardState(
            boxes_positions=[149, 130, 131, 169, 150, 168], pushers_positions=[]
        ),
        BoardState(
            boxes_positions=[149, 130, 131, 169, 168, 150], pushers_positions=[]
        ),
    ]


@pytest.fixture
def switched_boxes(goals_positions):
    return {
        Config.DEFAULT_PIECE_ID: goals_positions[Config.DEFAULT_PIECE_ID],
        Config.DEFAULT_PIECE_ID + 1: goals_positions[Config.DEFAULT_PIECE_ID + 1],
        Config.DEFAULT_PIECE_ID + 2: goals_positions[Config.DEFAULT_PIECE_ID + 2],
        Config.DEFAULT_PIECE_ID + 3: goals_positions[Config.DEFAULT_PIECE_ID + 3],
        Config.DEFAULT_PIECE_ID + 4: goals_positions[Config.DEFAULT_PIECE_ID + 4],
        Config.DEFAULT_PIECE_ID + 5: goals_positions[Config.DEFAULT_PIECE_ID + 5],
    }


@pytest.fixture
def switched_goals(boxes_positions):
    return {
        Config.DEFAULT_PIECE_ID: boxes_positions[Config.DEFAULT_PIECE_ID],
        Config.DEFAULT_PIECE_ID + 1: boxes_positions[Config.DEFAULT_PIECE_ID + 1],
        Config.DEFAULT_PIECE_ID + 2: boxes_positions[Config.DEFAULT_PIECE_ID + 2],
        Config.DEFAULT_PIECE_ID + 3: boxes_positions[Config.DEFAULT_PIECE_ID + 3],
        Config.DEFAULT_PIECE_ID + 4: boxes_positions[Config.DEFAULT_PIECE_ID + 4],
        Config.DEFAULT_PIECE_ID + 5: boxes_positions[Config.DEFAULT_PIECE_ID + 5],
    }


@pytest.fixture
def switched_goals_plus(boxes_positions):
    # boxorder 1 3 2
    # goalorder 3 2 1
    # (box, goal) id pairs [(2, 1), (3, 2), (1, 3), (4, 4), (5, 5), (6, 6)]
    return {
        Config.DEFAULT_PIECE_ID: boxes_positions[Config.DEFAULT_PIECE_ID + 1],
        Config.DEFAULT_PIECE_ID + 1: boxes_positions[Config.DEFAULT_PIECE_ID + 2],
        Config.DEFAULT_PIECE_ID + 2: boxes_positions[Config.DEFAULT_PIECE_ID],
        Config.DEFAULT_PIECE_ID + 3: boxes_positions[Config.DEFAULT_PIECE_ID + 3],
        Config.DEFAULT_PIECE_ID + 4: boxes_positions[Config.DEFAULT_PIECE_ID + 4],
        Config.DEFAULT_PIECE_ID + 5: boxes_positions[Config.DEFAULT_PIECE_ID + 5],
    }


@pytest.fixture
def switched_boxes_plus(goals_positions):
    # boxorder 1 3 2
    # goalorder 3 2 1
    # (box, goal) id pairs [(2, 1), (3, 2), (1, 3), (4, 4), (5, 5), (6, 6)]
    return {
        Config.DEFAULT_PIECE_ID: goals_positions[Config.DEFAULT_PIECE_ID + 2],
        Config.DEFAULT_PIECE_ID + 1: goals_positions[Config.DEFAULT_PIECE_ID],
        Config.DEFAULT_PIECE_ID + 2: goals_positions[Config.DEFAULT_PIECE_ID + 1],
        Config.DEFAULT_PIECE_ID + 3: goals_positions[Config.DEFAULT_PIECE_ID + 3],
        Config.DEFAULT_PIECE_ID + 4: goals_positions[Config.DEFAULT_PIECE_ID + 4],
        Config.DEFAULT_PIECE_ID + 5: goals_positions[Config.DEFAULT_PIECE_ID + 5],
    }


@pytest.fixture
def boxes_ids():
    return [
        Config.DEFAULT_PIECE_ID,
        Config.DEFAULT_PIECE_ID + 1,
        Config.DEFAULT_PIECE_ID + 2,
        Config.DEFAULT_PIECE_ID + 3,
        Config.DEFAULT_PIECE_ID + 4,
        Config.DEFAULT_PIECE_ID + 5,
    ]


@pytest.fixture
def goals_ids():
    return [
        Config.DEFAULT_PIECE_ID,
        Config.DEFAULT_PIECE_ID + 1,
        Config.DEFAULT_PIECE_ID + 2,
        Config.DEFAULT_PIECE_ID + 3,
        Config.DEFAULT_PIECE_ID + 4,
        Config.DEFAULT_PIECE_ID + 5,
    ]


@pytest.fixture
def pusher_ids():
    return [Config.DEFAULT_PIECE_ID, Config.DEFAULT_PIECE_ID + 1]


@pytest.fixture
def normalized_pushers_positions(board_width):
    return {
        Config.DEFAULT_PIECE_ID: index_1d(5, 1, board_width),
        Config.DEFAULT_PIECE_ID + 1: index_1d(8, 4, board_width),
    }


class DescribeBoardManager:
    def it_memoizes_pushers(
        self, board_graph, pushers_positions, invalid_pusher_position
    ):
        board_manager = BoardManager(board_graph)

        assert board_manager.pushers_count == 2
        assert sorted(board_manager.pushers_ids) == list(pushers_positions.keys())
        assert board_manager.pushers_positions == pushers_positions

        for pusher_id, pusher_position in pushers_positions.items():
            assert board_manager.pusher_position(pusher_id) == pusher_position
            assert board_manager.pusher_id_on(pusher_position) == pusher_id
            assert board_manager.has_pusher(pusher_id) is True
            assert board_manager.has_pusher_on(pusher_position) is True

        with pytest.raises(KeyError):
            board_manager.pusher_position(
                Config.DEFAULT_PIECE_ID + board_manager.pushers_count
            )
        with pytest.raises(KeyError):
            board_manager.pusher_id_on(invalid_pusher_position)

        assert (
            board_manager.has_pusher(
                Config.DEFAULT_PIECE_ID + board_manager.pushers_count
            )
            is False
        )
        assert board_manager.has_pusher_on(invalid_pusher_position) is False

    def it_memoizes_boxes(
        self, board_graph, boxes_positions, boxes_ids, invalid_box_position
    ):
        board_manager = BoardManager(board_graph)

        assert board_manager.boxes_count == 6
        assert sorted(board_manager.boxes_ids) == boxes_ids
        assert board_manager.boxes_positions == boxes_positions

        for box_id, box_position in boxes_positions.items():
            assert board_manager.box_position(box_id) == box_position
            assert board_manager.box_id_on(box_position) == box_id
            assert board_manager.has_box(box_id) is True
            assert board_manager.has_box_on(box_position) is True

        with pytest.raises(KeyError):
            board_manager.box_position(
                Config.DEFAULT_PIECE_ID + board_manager.boxes_count
            )
        with pytest.raises(KeyError):
            board_manager.box_id_on(invalid_box_position)

        assert (
            board_manager.has_box(Config.DEFAULT_PIECE_ID + board_manager.boxes_count)
            is False
        )
        assert board_manager.has_box_on(invalid_box_position) is False

        for box_id, box_position in boxes_positions.items():
            assert board_manager.box_plus_id(box_id) == SokobanPlus.DEFAULT_PLUS_ID

        # Doesn't rise KeyError if plus is disabled...
        assert (
            board_manager.box_plus_id(
                Config.DEFAULT_PIECE_ID + board_manager.boxes_count
            )
            == SokobanPlus.DEFAULT_PLUS_ID
        )

        board_manager.boxorder = "1 3 2"
        board_manager.goalorder = "3 1 2"
        board_manager.enable_sokoban_plus()

        assert board_manager.box_plus_id(Config.DEFAULT_PIECE_ID) == 1
        assert board_manager.box_plus_id(Config.DEFAULT_PIECE_ID + 1) == 3
        assert board_manager.box_plus_id(Config.DEFAULT_PIECE_ID + 2) == 2
        assert (
            board_manager.box_plus_id(Config.DEFAULT_PIECE_ID + 3)
            == SokobanPlus.DEFAULT_PLUS_ID
        )
        assert (
            board_manager.box_plus_id(Config.DEFAULT_PIECE_ID + 4)
            == SokobanPlus.DEFAULT_PLUS_ID
        )
        assert (
            board_manager.box_plus_id(Config.DEFAULT_PIECE_ID + 5)
            == SokobanPlus.DEFAULT_PLUS_ID
        )

        with pytest.raises(KeyError):
            assert board_manager.box_plus_id(
                Config.DEFAULT_PIECE_ID + board_manager.boxes_count
            )

    def it_memoizes_goals(
        self, board_graph, goals_positions, goals_ids, invalid_goal_position
    ):
        board_manager = BoardManager(board_graph)

        assert board_manager.goals_count == 6
        assert sorted(board_manager.goals_ids) == goals_ids
        assert board_manager.goals_positions == goals_positions

        for goal_id, goal_position in goals_positions.items():
            assert board_manager.goal_position(goal_id) == goal_position
            assert board_manager.goal_id_on(goal_position) == goal_id
            assert board_manager.has_goal(goal_id) is True
            assert board_manager.has_goal_on(goal_position) is True

        with pytest.raises(KeyError):
            board_manager.goal_position(
                Config.DEFAULT_PIECE_ID + board_manager.goals_count
            )
        with pytest.raises(KeyError):
            board_manager.goal_id_on(invalid_goal_position)

        assert (
            board_manager.has_goal(Config.DEFAULT_PIECE_ID + board_manager.goals_count)
            is False
        )
        assert board_manager.has_goal_on(invalid_goal_position) is False

        for goal_id, goal_position in goals_positions.items():
            assert board_manager.goal_plus_id(goal_id) == SokobanPlus.DEFAULT_PLUS_ID

        # Doesn't rise KeyError if plus is disabled...
        assert (
            board_manager.goal_plus_id(
                Config.DEFAULT_PIECE_ID + board_manager.goals_count
            )
            == SokobanPlus.DEFAULT_PLUS_ID
        )

        board_manager.boxorder = "1 3 2"
        board_manager.goalorder = "3 1 2"
        board_manager.enable_sokoban_plus()

        assert board_manager.goal_plus_id(Config.DEFAULT_PIECE_ID) == 3
        assert board_manager.goal_plus_id(Config.DEFAULT_PIECE_ID + 1) == 1
        assert board_manager.goal_plus_id(Config.DEFAULT_PIECE_ID + 2) == 2
        assert (
            board_manager.goal_plus_id(Config.DEFAULT_PIECE_ID + 3)
            == SokobanPlus.DEFAULT_PLUS_ID
        )
        assert (
            board_manager.goal_plus_id(Config.DEFAULT_PIECE_ID + 4)
            == SokobanPlus.DEFAULT_PLUS_ID
        )
        assert (
            board_manager.goal_plus_id(Config.DEFAULT_PIECE_ID + 5)
            == SokobanPlus.DEFAULT_PLUS_ID
        )

        with pytest.raises(KeyError):
            assert board_manager.goal_plus_id(
                Config.DEFAULT_PIECE_ID + board_manager.goals_count
            )

    def it_calculates_all_valid_board_solutions(
        self, board_graph, all_solutions, sokoban_plus_solutions
    ):
        board_manager = BoardManager(board_graph)

        assert list(board_manager.solutions()) == all_solutions

        board_manager.boxorder = "1 3 2"
        board_manager.goalorder = "3 2 1"
        board_manager.enable_sokoban_plus()

        assert list(board_manager.solutions()) == sokoban_plus_solutions

    def it_moves_boxes(self, board_graph):
        board_manager = BoardManager(board_graph)

        old_box_position = board_manager.box_position(Config.DEFAULT_PIECE_ID)
        board_manager.move_box(Config.DEFAULT_PIECE_ID, 0)
        assert board_manager.box_position(Config.DEFAULT_PIECE_ID) == 0
        assert board_manager.has_box_on(0)
        assert not board_manager.has_box_on(old_box_position)
        assert board_manager.board[0].has_box
        assert not board_manager.board[old_box_position].has_box
        board_manager.move_box(Config.DEFAULT_PIECE_ID, old_box_position)

        board_manager.move_box_from(old_box_position, 0)
        assert board_manager.box_position(Config.DEFAULT_PIECE_ID) == 0
        assert board_manager.has_box_on(0)
        assert not board_manager.has_box_on(old_box_position)
        assert board_manager.board[0].has_box
        assert not board_manager.board[old_box_position].has_box

    def it_moves_pushers(self, board_graph):
        board_manager = BoardManager(board_graph)

        old_pusher_position = board_manager.pusher_position(Config.DEFAULT_PIECE_ID)
        board_manager.move_pusher(Config.DEFAULT_PIECE_ID, 0)
        assert board_manager.pusher_position(Config.DEFAULT_PIECE_ID) == 0
        assert board_manager.has_pusher_on(0)
        assert not board_manager.has_pusher_on(old_pusher_position)
        assert board_manager.board[0].has_pusher
        assert not board_manager.board[old_pusher_position].has_pusher
        board_manager.move_pusher(Config.DEFAULT_PIECE_ID, old_pusher_position)

        board_manager.move_pusher_from(old_pusher_position, 0)
        assert board_manager.pusher_position(Config.DEFAULT_PIECE_ID) == 0
        assert board_manager.has_pusher_on(0)
        assert not board_manager.has_pusher_on(old_pusher_position)
        assert board_manager.board[0].has_pusher
        assert not board_manager.board[old_pusher_position].has_pusher

    def test_moving_box_onto_obstacle_raises_exception(
        self, board_graph, pushers_positions, boxes_positions
    ):
        wall_position = index_1d(4, 0, board_graph.board_width)
        board_manager = BoardManager(board_graph)
        box_id = Config.DEFAULT_PIECE_ID
        box_position = board_manager.box_position(box_id)

        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_box(box_id, boxes_positions[Config.DEFAULT_PIECE_ID + 1])
        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_box_from(
                box_position, boxes_positions[Config.DEFAULT_PIECE_ID + 1]
            )

        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_box(box_id, pushers_positions[Config.DEFAULT_PIECE_ID])
        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_box_from(
                box_position, pushers_positions[Config.DEFAULT_PIECE_ID]
            )

        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_box(box_id, wall_position)
        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_box_from(box_position, wall_position)

    def test_moving_pusher_onto_obstacle_raises_exception(
        self, board_graph, boxes_positions, pushers_positions
    ):
        wall_position = index_1d(4, 0, board_graph.board_width)
        board_manager = BoardManager(board_graph)
        pusher_id = Config.DEFAULT_PIECE_ID
        pusher_position = board_manager.pusher_position(pusher_id)

        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_pusher(
                pusher_id, boxes_positions[Config.DEFAULT_PIECE_ID]
            )
        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_pusher_from(
                pusher_position, boxes_positions[Config.DEFAULT_PIECE_ID]
            )

        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_pusher(
                pusher_id, pushers_positions[Config.DEFAULT_PIECE_ID + 1]
            )
        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_pusher_from(
                pusher_position, pushers_positions[Config.DEFAULT_PIECE_ID + 1]
            )

        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_pusher(pusher_id, wall_position)
        with pytest.raises(CellAlreadyOccupiedError):
            board_manager.move_pusher_from(pusher_position, wall_position)

    def it_implements_switching_box_and_goal_positions(
        self,
        board_graph,
        boxes_positions,
        switched_boxes,
        goals_positions,
        switched_goals,
        puzzle,
        switched_puzzle,
    ):
        board_manager = BoardManager(board_graph)

        board_manager.switch_boxes_and_goals()
        assert board_manager.boxes_positions == switched_boxes
        assert board_manager.goals_positions == switched_goals
        assert str(board_manager.board) == str(switched_puzzle)

        board_manager.switch_boxes_and_goals()
        assert board_manager.boxes_positions == boxes_positions
        assert board_manager.goals_positions == goals_positions
        assert str(board_manager.board) == str(puzzle)

    def test_switching_respects_sokoban_plus_if_enabled(
        self,
        board_graph,
        switched_boxes_plus,
        switched_goals_plus,
        boxes_positions,
        goals_positions,
    ):
        board_manager = BoardManager(board_graph, boxorder="1 3 2", goalorder="3 2 1")

        board_manager.enable_sokoban_plus()

        board_manager.switch_boxes_and_goals()
        assert board_manager.boxes_positions == switched_boxes_plus
        assert board_manager.goals_positions == switched_goals_plus

        board_manager.switch_boxes_and_goals()
        assert board_manager.boxes_positions == boxes_positions
        assert board_manager.goals_positions == goals_positions

    def test_switching_moves_pusher_out_of_the_way(self):
        board = "\n".join(
            [
                "########",
                "#------#",
                "#--+---#",
                "#----$-#",
                "########",
            ]
        )
        switched_board = "\n".join(
            [
                "########",
                "#------#",
                "#--$---#",
                "#----+-#",
                "########",
            ]
        )

        puzzle = SokobanPuzzle(board=board)
        graph = BoardGraph(puzzle)
        board_manager = BoardManager(graph)

        board_manager.switch_boxes_and_goals()
        assert str(board_manager.board) == switched_board
        board_manager.switch_boxes_and_goals()
        assert str(board_manager.board) == board



from __future__ import annotations

import enum
from dataclasses import dataclass
from itertools import groupby
from typing import Iterable, List, Optional

from .board_graph import BoardGraph
from .board_manager import CellAlreadyOccupiedError
from .config import Direction, Config
from .hashed_board_manager import HashedBoardManager
from .pusher_step import PusherStep


class SolvingMode(enum.Enum):
    FORWARD = 0
    REVERSE = 1

    def __repr__(self):
        return "SolvingMode." + self.name


class NonPlayableBoardError(RuntimeError):
    def __init__(self):
        super().__init__("Board is not playable!")


class IllegalMoveError(RuntimeError):
    pass


@dataclass
class MoveWorkerOptions:
    decrease_pull_count: Optional[int] = None
    increase_pull_count: Optional[int] = None
    force_pulls: Optional[bool] = None


class Mover:
    """Implements game rules (on-board movement).

    Supports forward and reverse solving mode.

    **Forward solving mode**

    - pusher is allowed to push single box at the time
    - pusher can't pull boxes
    - pusher can't jump over boxes or walls

    **Reverse solving mode:**

    - pusher is allowed to pull single box at the time

        - if position allows pull that pull is optional (pusher is allowed to move
          without pull even if pull is possible).
        - default behavior is to always pull boxes but that can be changed any time
          through `.pulls_boxes`

    - pusher can't push boxes

    - pusher is allowed to jump over boxes and walls

        - jumps are allowed only before first pull is done

    - board starts in solved state: positions of boxes and goals are switched

    **History management**

    Mover only stores last performed move in history and it doesn't offer redo.
    Failed moves, undo and non-moves (ie. selecting already selected pusher or
    jumping on same position pusher is already standing on) clear undo history.

    Warning:
        :class:`.Mover` operates directly on referenced game board. Because of that,
        this board should not be edited outside of :class:`.Mover` once
        :class:`.Mover` instance had been attached to it: editing the board will
        corrupt :class:`.Mover` internal state. For the same reason, it is not
        allowed to attach two movers to same game board.
    """

    def __init__(
        self, board: BoardGraph, solving_mode: SolvingMode = SolvingMode.FORWARD
    ):
        """
        Args:
            board: Instance of :class:`.BoardGraph` subclasses
            solving_mode: start the game in this solving mode
        """
        self._manager = HashedBoardManager(board)
        self._solving_mode = solving_mode
        self._pulls_boxes = True
        self._selected_pusher: int = Config.DEFAULT_PIECE_ID
        self._pull_count: int = 0
        self._last_move: List[PusherStep] = []

        if not self._manager.is_playable:
            raise NonPlayableBoardError

        if self._solving_mode == SolvingMode.REVERSE:
            self._manager.switch_boxes_and_goals()

    @property
    def board(self) -> BoardGraph:
        """Board on which :class:`.Mover` is operating on"""
        return self._manager.board

    @property
    def solving_mode(self) -> SolvingMode:
        """:class:`.Mover` operation mode (:class:`.SolvingMode`)."""
        return self._solving_mode

    @property
    def board_manager(self) -> HashedBoardManager:
        """Current board manager (:class:`.HashedBoardManager`)."""
        return self._manager

    @property
    def selected_pusher(self) -> int:
        """ID of pusher that will perform next move.

        See Also:
            :meth:`.select_pusher`
        """
        return self._selected_pusher

    @property
    def pulls_boxes(self) -> bool:
        """
        Select behavior in `.SolvingMode.REVERSE` mode when pusher is moving away
        from box.

        See Also:
            :meth:`.move`
        """
        return self._pulls_boxes

    @pulls_boxes.setter
    def pulls_boxes(self, rv: bool):
        self._pulls_boxes = rv

    @property
    def last_move(self) -> List[PusherStep]:
        """Sequence of :class:`.PusherStep` that contains most recent movement.

        Whenever :class:`.Mover` performs any movement or pusher selection, it puts
        resulting :class:`.PusherStep` into this sequence in order pusher steps
        happened.

        This is useful for movement animation in GUI. After Mover performs movement,
        GUI has enough information to know what was performed and to choose which
        animations to render for that.

        It is also possible to set this to some external sequence of moves. In that
        case, calling :meth:`.undo_last_move` will cause Mover to try to undo that
        external sequence of pusher steps.

        Example:

            >>> from sokoenginepy.game import Mover, PusherStep, Direction, BoardGraph
            >>> from sokoenginepy.io import SokobanPuzzle
            >>> puzzle = SokobanPuzzle(board='\\n'.join([
            ...     '    #####',
            ...     '    #  @#',
            ...     '    #$  #',
            ...     '  ###  $##',
            ...     '  #  $ $ #',
            ...     '### # ## #   ######',
            ...     '#   # ## #####  ..#',
            ...     '# $  $          ..#',
            ...     '##### ### #@##  ..#',
            ...     '    #     #########',
            ...     '    #######'
            ... ]))
            >>> board = BoardGraph(puzzle)
            >>> mover = Mover(board)
            >>> mover.last_move = [PusherStep(Direction.UP), PusherStep(Direction.RIGHT)]
            >>> mover.undo_last_move()
            >>> print(mover.board)
            ----#####----------
            ----#---#----------
            ----#$@-#----------
            --###--$##---------
            --#--$-$-#---------
            ###-#-##-#---######
            #---#-##-#####--..#
            #-$--$----------..#
            #####-###-#@##--..#
            ----#-----#########
            ----#######--------
            >>> mover.last_move
            [PusherStep(Direction.LEFT, box_moved=False), PusherStep(Direction.DOWN, box_moved=False)]

        Warning:
            Subsequent movement overwrites this meaning that Mover can only undo last
            move performed (it doesn't keep whole history of movement, only the last
            move).
        """
        return self._last_move

    @last_move.setter
    def last_move(self, rv: List[PusherStep]):
        self._last_move = rv

    def select_pusher(self, pusher_id: int):
        """
        Selects pusher that will perform next move.

        Mover always selects :data:`Config.DEFAULT_PIECE_ID` before any movements is
        performed. This means that for single-pusher boards, that single pusher is
        always automatically selected and this method doesn't need to be called.

        Args:
            pusher_id: ID of pusher

        See Also:
            `.BoardManager.pushers_ids`
        """

        if pusher_id == self._selected_pusher:
            return

        old_pusher_position = self._manager.pusher_position(self._selected_pusher)
        new_pusher_position = self._manager.pusher_position(pusher_id)
        selection_path = self._manager.board.positions_path_to_directions_path(
            self._manager.board.find_jump_path(old_pusher_position, new_pusher_position)
        )

        self._last_move = []
        for direction in selection_path:
            pusher_step = PusherStep(direction, False)
            pusher_step.is_pusher_selection = True
            self._last_move.append(pusher_step)

        self._selected_pusher = pusher_id

    def move(self, direction: Direction):
        """Moves currently selected pusher in ``direction``.

        In `.SolvingMode.FORWARD` mode, pushes the box in front of pusher (if there
        is one).

        In `.SolvingMode.REVERSE` mode pulls box together with pusher (if there is
        one and if ``self.pulls_boxes is True``).

        Args:
            direction: direction of movement

        Raises:
            IllegalMoveError: for illegal moves
        """
        options = MoveWorkerOptions()
        if self._solving_mode == SolvingMode.FORWARD:
            options.decrease_pull_count = False
            self._push_or_move(direction, options)
        else:
            options.force_pulls = self._pulls_boxes
            options.increase_pull_count = True
            self._pull_or_move(direction, options)

    def jump(self, new_position: int):
        """Currently selected pusher jumps to ``new_position``.

        Fails if

            - :class:`.Mover` is in `.SolvingMode.FORWARD` mode
            - pusher can't be dropped on ``new_position``
            - first pull had been made

        Raises:
            IllegalMoveError: for illegal jumps
        """
        if self._pull_count != 0:
            raise IllegalMoveError("Jumps not allowed after first pull")

        if self._solving_mode != SolvingMode.REVERSE:
            raise IllegalMoveError("Jumps allowed only in reverse solving mode")

        old_position = self._manager.pusher_position(self._selected_pusher)
        if old_position == new_position:
            return

        try:
            self._manager.move_pusher_from(old_position, new_position)
        except CellAlreadyOccupiedError as exc:
            raise IllegalMoveError(str(exc))

        path = self._manager.board.positions_path_to_directions_path(
            self._manager.board.find_jump_path(old_position, new_position)
        )

        def jump_am(direction):
            pusher_step = PusherStep(direction, False)
            pusher_step.is_jump = True
            pusher_step.pusher_id = self._selected_pusher
            return pusher_step

        self._last_move = [jump_am(direction) for direction in path]

    def undo_last_move(self):
        """
        Takes sequence of moves stored in self.last_move and undoes it.

        See Also:
            `.Mover.last_move`
        """
        new_last_moves = []
        old_last_moves = self._last_move

        jump_key = 0
        pusher_change_key = 1
        move_key = 2

        def key_functor(elem):
            if elem.is_jump:
                return jump_key
            if elem.is_pusher_selection:
                return pusher_change_key
            return move_key

        for moves_type, moves_group in groupby(reversed(old_last_moves), key_functor):
            if moves_type == move_key:
                for pusher_step in moves_group:
                    self._undo_pusher_step(pusher_step)
                    new_last_moves += self._last_move
            elif moves_type == jump_key:
                self._undo_jump(moves_group)
                new_last_moves += self._last_move
            else:
                self._undo_pusher_selection(moves_group)
                new_last_moves += self._last_move

        self._last_move = new_last_moves

    def _undo_pusher_step(self, pusher_step: PusherStep):
        options = MoveWorkerOptions()
        if self._solving_mode == SolvingMode.FORWARD:
            has_box_behind_pusher = self.board_manager.has_box_on(
                self._manager.board.neighbor(
                    self.board_manager.pusher_position(self.selected_pusher),
                    pusher_step.direction,
                )
            )

            if not pusher_step.is_move and not has_box_behind_pusher:
                raise IllegalMoveError("Requested push undo, but no box behind pusher!")
            options.force_pulls = not pusher_step.is_move
            options.increase_pull_count = False
            self._pull_or_move(pusher_step.direction.opposite, options)
        else:
            options.decrease_pull_count = True
            self._push_or_move(pusher_step.direction.opposite, options)

    def _undo_jump(self, jump_moves: Iterable[PusherStep]):
        path = [pusher_step.direction.opposite for pusher_step in jump_moves]
        old_position = self._manager.pusher_position(self._selected_pusher)
        new_position = self._manager.board.path_destination(old_position, path)
        self.jump(new_position)

    def _undo_pusher_selection(self, selection_moves: Iterable[PusherStep]):
        path = [pusher_step.direction.opposite for pusher_step in selection_moves]
        old_position = self._manager.pusher_position(self._selected_pusher)
        new_position = self._manager.board.path_destination(old_position, path)
        self.select_pusher(self._manager.pusher_id_on(new_position))

    def _push_or_move(self, direction: Direction, options: MoveWorkerOptions):
        """
        Perform movement of currently selected pusher in ``direction``.

        In case there is a box in front of pusher, pushes it.
        """
        initial_pusher_position = self._manager.pusher_position(self._selected_pusher)
        in_front_of_pusher = self._manager.board.neighbor(
            initial_pusher_position, direction
        )

        if not in_front_of_pusher:
            raise IllegalMoveError(
                "Can't move pusher off board! (ID: "
                + "{0}, direction: {1})".format(self._selected_pusher, str(direction))
            )

        is_push = False
        in_front_of_box = None
        if self._manager.has_box_on(in_front_of_pusher):
            is_push = True
            in_front_of_box = self._manager.board.neighbor(
                in_front_of_pusher, direction
            )
            if not in_front_of_box:
                raise IllegalMoveError(
                    "Can't push box off board! (ID: "
                    + "{0}, direction: {1})".format(
                        self._manager.box_id_on(in_front_of_pusher), str(direction)
                    )
                )

            try:
                self._manager.move_box_from(in_front_of_pusher, in_front_of_box)
            except CellAlreadyOccupiedError as exc:
                raise IllegalMoveError(str(exc))

        try:
            self._manager.move_pusher_from(initial_pusher_position, in_front_of_pusher)
        except CellAlreadyOccupiedError as exc:
            raise IllegalMoveError(str(exc))

        pusher_step = PusherStep(direction, is_push)
        pusher_step.pusher_id = self._selected_pusher
        if is_push:
            pusher_step.moved_box_id = self._manager.box_id_on(in_front_of_box)
            if options.decrease_pull_count and self._pull_count > 0:
                self._pull_count -= 1
        self._last_move = [pusher_step]

    def _pull_or_move(self, direction: Direction, options: MoveWorkerOptions):
        """
        Perform movement of currently selected pusher in ``direction``.

        In case there is a box in behind of pusher, might pull it.
        """
        initial_pusher_position = self._manager.pusher_position(self._selected_pusher)
        in_front_of_pusher = self._manager.board.neighbor(
            initial_pusher_position, direction
        )

        if not in_front_of_pusher:
            raise IllegalMoveError(
                "Can't move pusher off board! (ID: "
                + "{0}, direction: {1})".format(self._selected_pusher, str(direction))
            )

        try:
            self._manager.move_pusher_from(initial_pusher_position, in_front_of_pusher)
        except CellAlreadyOccupiedError as exc:
            raise IllegalMoveError(str(exc))

        is_pull = False
        if options.force_pulls:
            behind_pusher = self._manager.board.neighbor(
                initial_pusher_position, direction.opposite
            )
            if behind_pusher and self._manager.board[behind_pusher].has_box:
                is_pull = True
                try:
                    self._manager.move_box_from(behind_pusher, initial_pusher_position)
                except CellAlreadyOccupiedError as exc:
                    raise IllegalMoveError(str(exc))
                if options.increase_pull_count:
                    self._pull_count += 1

        pusher_step = PusherStep(direction, is_pull)
        pusher_step.pusher_id = self._selected_pusher
        if is_pull:
            pusher_step.moved_box_id = self._manager.box_id_on(initial_pusher_position)
        self._last_move = [pusher_step]



import itertools
import re
from functools import reduce
from typing import Iterable, List, Optional, Union

_RE_ONLY_DIGITS_AND_SPACES = re.compile(r"^([0-9\s])*$")


def is_blank(data: Optional[Union[str, List[str]]]) -> bool:
    """True if ``line`` is empty or it contains only spaces."""

    _check = lambda l: l is None or l.strip() == ""

    if not data:
        return True

    if isinstance(data, str):
        return _check(data)

    return all(_check(_) for _ in data)


def contains_only_digits_and_spaces(line: Optional[str]) -> bool:
    return reduce(
        lambda x, y: x and y,
        [
            True if _RE_ONLY_DIGITS_AND_SPACES.match(l) else False
            for l in line.splitlines()
        ],
        True,
    )


def in_batches(iterable, of_size=1):
    """
    Generator that yields generator slices of iterable.

    Since it is elegant and working flawlessly, it is shameles C/P from
    https://stackoverflow.com/questions/8991506/iterate-an-iterator-by-chunks-of-n-in-python/8998040#8998040

    Warning:
        Each returned batch should be completely consumed before next batch
        is yielded. See example below to better understand what that means.

    Example::

        g = (o for o in range(10))
        for batch in in_batches(g, of_size=3):
            print(list(batch))
        # [0, 1, 2]
        # [3, 4, 5]
        # [6, 7, 8]
        # [9]

        # And don't consume whole batch before yielding another one...
        g = list(range(10))
        for batch in in_batches(g, of_size=3):
            print( [next(batch), next(batch)] )
        # [0, 1]
        # [2, 3]
        # [4, 5]
        # [6, 7]
        # [8, 9]
    """
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, of_size)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)



from __future__ import annotations

from typing import TYPE_CHECKING, Final, List, Optional

from .rle import Rle
from .utilities import contains_only_digits_and_spaces, is_blank

if TYPE_CHECKING:
    from ..game import BaseTessellation, PusherStep, Tessellation
    from .snapshot_parsing import MovementTokens


class Snapshot:
    """
    Base class for game snapshots and accompanying metadata.

    Game snapshot is sequence of pusher steps representing actual steps, jumps (in
    reverse solving mode) and pusher selections (in Multiban variant).
    """

    # Characters used in textual representation of game snapshot. Not all game variants
    # use all characters. Also, for different game variants, same character may have
    # different meaning (represent different direction of movement).

    l: Final[str] = "l"
    u: Final[str] = "u"
    r: Final[str] = "r"
    d: Final[str] = "d"
    L: Final[str] = "L"
    U: Final[str] = "U"
    R: Final[str] = "R"
    D: Final[str] = "D"
    w: Final[str] = "w"
    W: Final[str] = "W"
    e: Final[str] = "e"
    E: Final[str] = "E"
    n: Final[str] = "n"
    N: Final[str] = "N"
    s: Final[str] = "s"
    S: Final[str] = "S"

    JUMP_BEGIN: Final[str] = "["
    JUMP_END: Final[str] = "]"
    PUSHER_CHANGE_BEGIN: Final[str] = "{"
    PUSHER_CHANGE_END: Final[str] = "}"
    CURRENT_POSITION_CH: Final[str] = "*"

    @classmethod
    def is_move_step(cls, character: str) -> bool:
        from .snapshot_parsing import Constants

        return character in Constants.MOVE_CHARACTERS

    @classmethod
    def is_push_step(cls, character: str) -> bool:
        from .snapshot_parsing import Constants

        return character in Constants.PUSH_CHARACTERS

    @classmethod
    def is_pusher_step(cls, character: str) -> bool:
        return cls.is_move_step(character) or cls.is_push_step(character)

    @classmethod
    def is_marker(cls, character: str) -> bool:
        from .snapshot_parsing import Constants

        return character in Constants.MARKERS

    @classmethod
    def is_snapshot(cls, line: str) -> bool:
        """
        True if ``line`` contains only:

        - movement characters
        - other snapshot characters (ie. jump markers)
        - Rle characters
        - spaces and newlines
        """
        from .snapshot_parsing import Constants

        return (
            not is_blank(line)
            and not contains_only_digits_and_spaces(line)
            and all(
                True if Constants.RE_SNAPSHOT_STRING.match(l) else False
                for l in line.splitlines()
            )
        )

    def __init__(self, tessellation: Tessellation, moves_data: str = ""):
        self.title: str = ""
        self.solver: str = ""
        self.notes: List[str] = []

        self._tessellation = tessellation
        self._tessellation_obj_val: Optional[BaseTessellation] = None

        if not is_blank(moves_data) and not self.is_snapshot(moves_data):
            raise ValueError("Invalid characters in snapshot string!")
        self._moves_data: str = moves_data or ""

        self._parsed_moves: MovementTokens = []
        self._was_parsed = False
        self._pushes_count: int = 0
        self._moves_count: int = 0
        self._jumps_count: int = 0
        self._is_reverse: bool = False

    def to_str(self, rle_encode=False) -> str:
        self._reparse_if_not_parsed()

        retv = "".join(str(_) for _ in self._parsed_moves)

        # Reverse snapshots must start with jump, even if it is empty one
        if self.is_reverse and (
            not self._parsed_moves
            or (self._parsed_moves and retv[0] != self.JUMP_BEGIN)
        ):
            retv = self.JUMP_BEGIN + self.JUMP_END + retv

        if rle_encode:
            retv = Rle.encode(retv)

        return retv

    def __str__(self):
        return self.to_str(rle_encode=False)

    def __repr__(self):
        klass = self.__class__.__name__
        return f'{klass}(moves_data="{self.to_str(rle_encode=False)}")'

    @property
    def tessellation(self) -> Tessellation:
        return self._tessellation

    @property
    def _tessellation_obj(self):
        from ..game import BaseTessellation

        if self._tessellation_obj_val is None:
            self._tessellation_obj_val = BaseTessellation.instance(self._tessellation)

        return self._tessellation_obj_val

    @property
    def moves_data(self) -> str:
        return self._moves_data

    @moves_data.setter
    def moves_data(self, rv: str):
        if not is_blank(rv) and not self.is_snapshot(rv):
            raise ValueError("Invalid characters in snapshot string!")
        self._moves_data = rv or ""
        self._was_parsed = False

    @property
    def pusher_steps(self) -> List[PusherStep]:
        """
        Game engine representation of pusher movement.
        """
        self._reparse_if_not_parsed()
        return sum(
            (_.pusher_steps(self._tessellation_obj) for _ in self._parsed_moves), []
        )

    @pusher_steps.setter
    def pusher_steps(self, rv: List[PusherStep]):
        """
        Warning:
            This will replace ``moves_data`` with ``rv`` converted to ``str``.
        """
        from .snapshot_parsing import Jump, PusherSelection, Steps

        i = 0
        iend = len(rv)

        self._parsed_moves = []
        self._pushes_count = 0
        self._moves_count = 0
        self._jumps_count = 0
        self._is_reverse = False
        self._was_parsed = True

        while i < iend:
            if rv[i].is_jump:
                jump = Jump("")
                while i < iend and rv[i].is_jump:
                    jump.data += self._tessellation_obj.pusher_step_to_char(rv[i])
                    if rv[i].is_current_pos:
                        jump.data += self.CURRENT_POSITION_CH
                    i += 1
                self._parsed_moves.append(jump)
                self._jumps_count += 1
                self._is_reverse = True
                self._moves_count += jump.moves_count

            elif rv[i].is_pusher_selection:
                pusher_selection = PusherSelection("")
                while i < iend and rv[i].is_pusher_selection:
                    pusher_selection.data += self._tessellation_obj.pusher_step_to_char(
                        rv[i]
                    )
                    if rv[i].is_current_pos:
                        pusher_selection.data += self.CURRENT_POSITION_CH
                    i += 1
                self._parsed_moves.append(pusher_selection)

            else:
                steps = Steps("")
                while i < iend and not rv[i].is_jump and not rv[i].is_pusher_selection:
                    steps.data += self._tessellation_obj.pusher_step_to_char(rv[i])
                    if rv[i].is_current_pos:
                        steps.data += self.CURRENT_POSITION_CH
                    i += 1
                self._parsed_moves.append(steps)
                self._pushes_count += steps.pushes_count
                self._moves_count += steps.moves_count

        self._moves_data = self.to_str(rle_encode=False)

    @property
    def pushes_count(self) -> int:
        """
        Count of box pushing steps.
        """
        self._reparse_if_not_parsed()
        return self._pushes_count

    @property
    def moves_count(self) -> int:
        """
        Count of steps that are:
            - not pushing a box
            - jumping (in reverse solving mode)
        """
        self._reparse_if_not_parsed()
        return self._moves_count

    @property
    def jumps_count(self) -> int:
        """
        Count of sequences of steps that are jumps. Jumps are possible when board is
        being solved in reverse mode.
        """
        self._reparse_if_not_parsed()
        return self._jumps_count

    @property
    def is_reverse(self) -> bool:
        """
        True if snapshot contains any jumps.
        """
        self._reparse_if_not_parsed()
        return self._is_reverse

    def _reparse_if_not_parsed(self):
        if not self._was_parsed:
            self._reparse()

    def _reparse(self):
        from .snapshot_parsing import Jump, Parser

        self._parsed_moves = []
        self._pushes_count = 0
        self._moves_count = 0
        self._jumps_count = 0
        self._is_reverse = False

        self._parsed_moves = Parser.parse(self._moves_data)

        for _ in self._parsed_moves:
            if isinstance(_, Jump):
                self._jumps_count += 1
                self._is_reverse = True
            self._moves_count += _.moves_count
            self._pushes_count += _.pushes_count

        self._was_parsed = True



from typing import Final, List, Optional


class BoardState:
    """
    Sample of board state.
    """

    #: Integer used for situations where board has has not been calculated.
    NO_HASH: Final[int] = 0

    def __init__(
        self,
        pushers_positions: Optional[List[int]] = None,
        boxes_positions: Optional[List[int]] = None,
        zobrist_hash: int = NO_HASH,
    ):
        self.pushers_positions = pushers_positions or []
        self.boxes_positions = boxes_positions or []
        self.zobrist_hash = zobrist_hash or self.NO_HASH

    @property
    def pushers_positions(self) -> List[int]:
        """
        Positions of pushers sorted by pusher ID.
        """
        return self._pushers_positions

    @pushers_positions.setter
    def pushers_positions(self, rv: List[int]):
        self._pushers_positions = rv or []

    @property
    def boxes_positions(self) -> List[int]:
        """
        Positions of boxes sorted by box ID.
        """
        return self._boxes_positions

    @boxes_positions.setter
    def boxes_positions(self, rv: List[int]):
        self._boxes_positions = rv or []

    @property
    def zobrist_hash(self) -> int:
        """
        Zobrist hash of state (see `.HashedBoardManager`)
        """
        return self._zobrist_hash

    @zobrist_hash.setter
    def zobrist_hash(self, rv: int):
        self._zobrist_hash = rv or self.NO_HASH

    def __str__(self) -> str:
        prefix = (len(self.__class__.__name__) + 2) * " "
        return "\n".join(
            [
                "<{} pushers_positions: {},".format(
                    self.__class__.__name__, self.pushers_positions
                ),
                prefix + "boxes_positions: {},".format(self.boxes_positions),
                prefix + "zobrist_hash: {}".format(self.zobrist_hash) + ">",
            ]
        )

    def __repr__(self) -> str:
        return "{}(pushers_positions={}, boxes_positions={}, zobrist_hash={})".format(
            self.__class__.__name__,
            repr(self.pushers_positions),
            repr(self.boxes_positions),
            repr(self.zobrist_hash),
        )

    def __eq__(self, other) -> bool:
        return (
            self.zobrist_hash is not None
            and self.zobrist_hash is not self.NO_HASH
            and self.zobrist_hash == other.zobrist_hash
        ) or (
            self.boxes_positions == other.boxes_positions
            and self.pushers_positions == other.pushers_positions
        )



from __future__ import annotations

import enum
import re
import textwrap
from functools import reduce
from operator import add
from typing import TYPE_CHECKING, Final, List, Optional, Set, Type

from .rle import Rle
from .snapshot import Snapshot
from .utilities import contains_only_digits_and_spaces, is_blank

if TYPE_CHECKING:
    from ..game import BaseTessellation, Tessellation
    from .puzzle_parsing import PuzzleParser, PuzzlePrinter, PuzzleResizer


class CellOrientation(enum.Enum):
    """
    Dynamic board cell property that depends on cell position in some tessellations.
    ie. in Trioban, origin of coordinate system is triangle pointing upwards. This means
    that orientation of all other triangles depends on orientation of origin.
    """

    DEFAULT = 0
    TRIANGLE_DOWN = 1
    OCTAGON = 2


class Puzzle:
    """
    Base class for game puzzles.

    Game puzzle is representation of game board together with all of its meta data and
    snapshots.

    It implements:
        - parsing board data from text
        - editing board: setting individual cells, resizing, trimming, ...

    All positions used are 1D array indexes.

    To convert 2D board coordinates into 1D array indexes, use ``game.index_1d``.
    To convert 1D array indexes into board 2D coordinates, use one of ``game.ROW``,
    ``game.x`` ``game.COLUMN`` and ``game.y``.
    """

    WALL: Final[str] = "#"
    PUSHER: Final[str] = "@"
    PUSHER_ON_GOAL: Final[str] = "+"
    BOX: Final[str] = "$"
    BOX_ON_GOAL: Final[str] = "*"
    GOAL: Final[str] = "."
    FLOOR: Final[str] = " "
    VISIBLE_FLOOR: Final[str] = "-"
    ALT_PUSHER1: Final[str] = "p"
    ALT_PUSHER2: Final[str] = "m"
    ALT_PUSHER_ON_GOAL1: Final[str] = "P"
    ALT_PUSHER_ON_GOAL2: Final[str] = "M"
    ALT_BOX1: Final[str] = "b"
    ALT_BOX_ON_GOAL1: Final[str] = "B"
    ALT_GOAL1: Final[str] = "o"
    ALT_VISIBLE_FLOOR1: Final[str] = "_"

    @classmethod
    def is_pusher(cls, character: str) -> bool:
        return (
            character == cls.PUSHER
            or character == cls.ALT_PUSHER1
            or character == cls.ALT_PUSHER2
            or character == cls.PUSHER_ON_GOAL
            or character == cls.ALT_PUSHER_ON_GOAL1
            or character == cls.ALT_PUSHER_ON_GOAL2
        )

    @classmethod
    def is_box(cls, character: str) -> bool:
        return (
            character == cls.BOX
            or character == cls.ALT_BOX1
            or character == cls.BOX_ON_GOAL
            or character == cls.ALT_BOX_ON_GOAL1
        )

    @classmethod
    def is_goal(cls, character: str) -> bool:
        return (
            character == cls.GOAL
            or character == cls.ALT_GOAL1
            or character == cls.BOX_ON_GOAL
            or character == cls.ALT_BOX_ON_GOAL1
            or character == cls.PUSHER_ON_GOAL
            or character == cls.ALT_PUSHER_ON_GOAL1
            or character == cls.ALT_PUSHER_ON_GOAL2
        )

    @classmethod
    def is_empty_floor(cls, character: str) -> bool:
        return (
            character == cls.FLOOR
            or character == cls.VISIBLE_FLOOR
            or character == cls.ALT_VISIBLE_FLOOR1
        )

    @classmethod
    def is_wall(cls, character: str) -> bool:
        return character == cls.WALL

    @classmethod
    def is_border_element(cls, character: str) -> bool:
        return (
            character == cls.WALL
            or character == cls.BOX_ON_GOAL
            or character == cls.ALT_BOX_ON_GOAL1
        )

    @classmethod
    def is_puzzle_element(cls, character: str) -> bool:
        return (
            cls.is_empty_floor(character)
            or cls.is_wall(character)
            or cls.is_pusher(character)
            or cls.is_box(character)
            or cls.is_goal(character)
        )

    @classmethod
    def is_board(cls, line: Optional[str]) -> bool:
        """
        Checks if line contains only characters legal in textual representation of
        boards.

        Note:
            Doesn't check if it actually contains legal board, it only checks that
            there are no illegal characters.
        """
        return not contains_only_digits_and_spaces(line) and reduce(
            lambda x, y: x and y,
            [True if _RE_BOARD_STRING.match(l) else False for l in line.splitlines()],
            True,
        )

    @classmethod
    def is_sokoban_plus(cls, line: str) -> bool:
        return contains_only_digits_and_spaces(line) and not is_blank(line)

    @classmethod
    def instance_from(
        cls,
        tessellation: Tessellation,
        width: int = 0,
        height: int = 0,
        board: Optional[str] = None,
    ) -> Puzzle:
        """
        Factory method. Produces instance of one of the subclasses.
        """
        from ..game import BaseTessellation
        from .hexoban import HexobanPuzzle
        from .octoban import OctobanPuzzle
        from .sokoban import SokobanPuzzle
        from .trioban import TriobanPuzzle

        tessellation_instance = BaseTessellation.instance(tessellation)

        for klass in [HexobanPuzzle, OctobanPuzzle, SokobanPuzzle, TriobanPuzzle]:
            if (
                tessellation_instance.__class__.__name__.replace(
                    "Tessellation", ""
                ).lower()
                in klass.__name__.lower()
            ):
                return klass(width=width, height=height, board=board)

        raise ValueError(tessellation)

    def __init__(
        self,
        tessellation: Tessellation,
        width: int = 0,
        height: int = 0,
        board: Optional[str] = None,
        resizer_cls: Optional[Type[PuzzleResizer]] = None,
        parser_cls: Optional[Type[PuzzleParser]] = None,
        printer_cls: Optional[Type[PuzzlePrinter]] = None,
    ):
        from .puzzle_parsing import PuzzleParser, PuzzlePrinter, PuzzleResizer

        self.title = ""
        self.author = ""
        self.boxorder = ""
        self.goalorder = ""
        self.notes: List[str] = []
        self.snapshots: List[Snapshot] = []

        self._pushers_count: Optional[int] = None
        self._boxes_count: Optional[int] = None
        self._goals_count: Optional[int] = None

        self._tessellation: Tessellation = tessellation
        self._tessellation_obj_val: Optional[BaseTessellation] = None
        self._resizer_cls: Type[PuzzleResizer] = resizer_cls or PuzzleResizer
        self._parser_cls: Type[PuzzleParser] = parser_cls or PuzzleParser
        self._printer_cls: Type[PuzzlePrinter] = printer_cls or PuzzlePrinter

        self._width: int
        self._height: int
        self._was_parsed: bool
        self._original_board: str
        # not str but list of single character strings. str is immutable and we need to
        # be able to modify individual board cells.
        self._parsed_board: List[str]

        if is_blank(board) or board is None:
            self._width = width
            self._height = height
            if self._width < 0 or self._height < 0:
                raise ValueError("Board dimensions can't be less than zero!")
            self._was_parsed = True
            self._original_board = ""
            self._parsed_board = width * height * [self.VISIBLE_FLOOR]

        else:
            self._width = 0
            self._height = 0
            self._was_parsed = False
            self._original_board = board
            self._parsed_board = []
            if not self.is_board(self._original_board):
                raise ValueError("Invalid characters in board string!")

    @property
    def tessellation(self) -> Tessellation:
        return self._tessellation

    @property
    def _tessellation_obj(self):
        from ..game import BaseTessellation

        if self._tessellation_obj_val is None:
            self._tessellation_obj_val = BaseTessellation.instance(self._tessellation)

        return self._tessellation_obj_val

    def cell_orientation(self, pos: int) -> CellOrientation:
        self._reparse_if_not_parsed()
        return self._tessellation_obj.cell_orientation(pos, self.width, self.height)

    def __getitem__(self, position: int) -> str:
        self._reparse_if_not_parsed()
        return self._parsed_board[position]

    def __setitem__(self, position: int, c: str):
        self._reparse_if_not_parsed()
        if not self.is_puzzle_element(c):
            raise ValueError("Invalid characters in board string!")
        self._parsed_board[position] = c

    def __contains__(self, position: int):
        self._reparse_if_not_parsed()
        return position < len(self._parsed_board)

    def __str__(self):
        return self.to_board_str(use_visible_floor=True)

    def __repr__(self):
        return "{klass}(board='\\n'.join([\n{board}\n]))".format(
            klass=self.__class__.__name__,
            board=textwrap.indent(
                ",\n".join(["'{0}'".format(l) for l in str(self).split("\n")]), "    "
            ),
        )

    def to_board_str(self, use_visible_floor=False, rle_encode=False) -> str:
        self._reparse_if_not_parsed()
        return self._printer_cls().print(
            self._parsed_board, self.width, self.height, use_visible_floor, rle_encode
        )

    @property
    def board(self) -> str:
        return self._original_board

    @board.setter
    def board(self, rv: str):
        if not self.is_board(rv):
            raise ValueError("Invalid characters in board string!")
        self._board = rv
        self._was_parsed = False
        self._pushers_count = None
        self._boxes_count = None
        self._goals_count = None

    @property
    def internal_board(self) -> str:
        """
        Internal, parsed board. For debugging purposes.
        """
        self._reparse_if_not_parsed()
        return "".join(self._parsed_board)

    @property
    def width(self) -> int:
        self._reparse_if_not_parsed()
        return self._width

    @property
    def height(self) -> int:
        self._reparse_if_not_parsed()
        return self._height

    @property
    def size(self) -> int:
        return self.width * self.height

    @property
    def pushers_count(self) -> int:
        if self._pushers_count is None:
            self._pushers_count = reduce(
                add, [1 if self.is_pusher(chr) else 0 for chr in self.board], 0
            )
        return self._pushers_count

    @property
    def boxes_count(self) -> int:
        if self._boxes_count is None:
            self._boxes_count = reduce(
                add, [1 if self.is_box(chr) else 0 for chr in self.board], 0
            )

        return self._boxes_count

    @property
    def goals_count(self) -> int:
        if self._goals_count is None:
            self._goals_count = reduce(
                add, [1 if self.is_goal(chr) else 0 for chr in self.board], 0
            )
        return self._goals_count

    def add_row_top(self):
        self._reparse_if_not_parsed()
        self._parsed_board, self._width, self._height = self._resizer.add_row_top(
            self._parsed_board, self.width, self.height
        )

    def add_row_bottom(self):
        self._reparse_if_not_parsed()
        (
            self._parsed_board,
            self._width,
            self._height,
        ) = self._resizer.add_row_bottom(self._parsed_board, self.width, self.height)

    def add_column_left(self):
        self._reparse_if_not_parsed()
        (
            self._parsed_board,
            self._width,
            self._height,
        ) = self._resizer.add_column_left(self._parsed_board, self.width, self.height)

    def add_column_right(self):
        self._reparse_if_not_parsed()
        (
            self._parsed_board,
            self._width,
            self._height,
        ) = self._resizer.add_column_right(self._parsed_board, self.width, self.height)

    def remove_row_top(self):
        self._reparse_if_not_parsed()
        (
            self._parsed_board,
            self._width,
            self._height,
        ) = self._resizer.remove_row_top(self._parsed_board, self.width, self.height)

    def remove_row_bottom(self):
        self._reparse_if_not_parsed()
        (
            self._parsed_board,
            self._width,
            self._height,
        ) = self._resizer.remove_row_bottom(self._parsed_board, self.width, self.height)

    def remove_column_left(self):
        self._reparse_if_not_parsed()
        (
            self._parsed_board,
            self._width,
            self._height,
        ) = self._resizer.remove_column_left(
            self._parsed_board, self.width, self.height
        )

    def remove_column_right(self):
        self._reparse_if_not_parsed()
        (
            self._parsed_board,
            self._width,
            self._height,
        ) = self._resizer.remove_column_right(
            self._parsed_board, self.width, self.height
        )

    def trim_left(self):
        self._reparse_if_not_parsed()
        self._parsed_board, self._width, self._height = self._resizer.trim_left(
            self._parsed_board, self.width, self.height
        )

    def trim_right(self):
        self._reparse_if_not_parsed()
        self._parsed_board, self._width, self._height = self._resizer.trim_right(
            self._parsed_board, self.width, self.height
        )

    def trim_top(self):
        self._reparse_if_not_parsed()
        self._parsed_board, self._width, self._height = self._resizer.trim_top(
            self._parsed_board, self.width, self.height
        )

    def trim_bottom(self):
        self._reparse_if_not_parsed()
        self._parsed_board, self._width, self._height = self._resizer.trim_bottom(
            self._parsed_board, self.width, self.height
        )

    def reverse_rows(self):
        self._reparse_if_not_parsed()
        (
            self._parsed_board,
            self._width,
            self._height,
        ) = self._resizer.reverse_rows(self._parsed_board, self.width, self.height)

    def reverse_columns(self):
        self._reparse_if_not_parsed()
        (
            self._parsed_board,
            self._width,
            self._height,
        ) = self._resizer.reverse_columns(self._parsed_board, self.width, self.height)

    def resize(self, new_width: int, new_height: int):
        """
        In-place resizing of board.

        Adds or removes rows and columns.
        """
        self._reparse_if_not_parsed()
        old_width = self.width
        old_height = self.height

        if new_height != old_height:
            if new_height > old_height:
                amount = new_height - old_height
                for _ in range(0, amount):
                    self.add_row_bottom()
            else:
                amount = old_height - new_height
                for _ in range(0, amount):
                    self.remove_row_bottom()

        if new_width != old_width:
            if new_width > old_width:
                amount = new_width - old_width
                for _ in range(0, amount):
                    self.add_column_right()
            else:
                amount = old_width - new_width
                for _ in range(0, amount):
                    self.remove_column_right()

    def resize_and_center(self, new_width: int, new_height: int):
        """
        In-place resizing of board.

        Adds or removes rows and columns keeping existing board centered inside of new
        one.
        """
        self._reparse_if_not_parsed()
        left = right = top = bottom = 0

        if new_width > self.width:
            left = int((new_width - self.width) / 2)
            right = new_width - self.width - left

        if new_height > self.height:
            top = int((new_height - self.height) / 2)
            bottom = new_height - self.height - top

        if (left, right, top, bottom) != (0, 0, 0, 0):
            for _ in range(0, left):
                self.add_column_left()
            for _ in range(0, top):
                self.add_row_top()

            if (right, bottom) != (0, 0):
                self.resize(self.width + right, self.height + bottom)

    def trim(self):
        """
        In-place resizing of board.

        Removes outer, blank rows and columns.
        """
        self.trim_top()
        self.trim_bottom()
        self.trim_left()
        self.trim_right()

    def _reparse(self):
        if not is_blank(self._original_board):
            board_rows = self._parser_cls().parse(self._original_board)
            self._height = len(board_rows)
            self._width = len(board_rows[0]) if self._height else 0
            self._parsed_board = sum((list(_) for _ in board_rows), [])
            for idx, val in enumerate(self._parsed_board):
                if val == self.FLOOR:
                    self._parsed_board[idx] = self.VISIBLE_FLOOR

        self._was_parsed = True

    def _reparse_if_not_parsed(self):
        if not self._was_parsed:
            self._reparse()

    @property
    def _resizer(self) -> PuzzleResizer:
        return self._resizer_cls()


_CHARACTERS: Set[str] = {
    Puzzle.WALL,
    Puzzle.PUSHER,
    Puzzle.PUSHER_ON_GOAL,
    Puzzle.BOX,
    Puzzle.BOX_ON_GOAL,
    Puzzle.GOAL,
    Puzzle.FLOOR,
    Puzzle.VISIBLE_FLOOR,
    Puzzle.ALT_PUSHER1,
    Puzzle.ALT_PUSHER2,
    Puzzle.ALT_PUSHER_ON_GOAL1,
    Puzzle.ALT_PUSHER_ON_GOAL2,
    Puzzle.ALT_BOX1,
    Puzzle.ALT_BOX_ON_GOAL1,
    Puzzle.ALT_GOAL1,
    Puzzle.ALT_VISIBLE_FLOOR1,
}
_RE_BOARD_STRING = re.compile(
    r"^([0-9\s"
    + re.escape("".join(_CHARACTERS))
    + re.escape("".join({Rle.GROUP_START, Rle.GROUP_END, Rle.EOL}))
    + "])*$"
)



from typing import Optional

from .config import Config, Direction


class PusherStep:
    """
    Represents single step of pusher movement.

    - ``move`` - pusher moved without pushing box
    - ``push/pull`` - pusher moved and pushed a box in front of it (
      :attr:`.SolvingMode.FORWARD`) or pusher moved and pulled a box behind it
      (:attr:`.SolvingMode.REVERSE`)
    - ``jump`` - single step in pusher jump sequence (jumps are allowed only in
      :attr:`.SolvingMode.REVERSE`)
    - ``pusher selection`` - single step in sequence describing focus change of active
      pusher in Multiban games
    """

    __slots__ = [
        "_box_moved",
        "_pusher_selected",
        "_pusher_jumped",
        "_pusher_id",
        "_moved_box_id",
        "direction",
        "is_current_pos",
    ]

    def __init__(
        self,
        direction: Direction = Direction.LEFT,
        box_moved: bool = False,
        is_jump: bool = False,
        is_pusher_selection: bool = False,
        pusher_id: int = Config.DEFAULT_PIECE_ID,
        moved_box_id: Optional[int] = None,
        is_current_pos: bool = False,
    ):
        if (box_moved or moved_box_id) and is_pusher_selection and is_jump:
            raise ValueError(
                "PusherStep can't be all, a push, a jump and a pusher selection!"
            )

        if is_jump and is_pusher_selection:
            raise ValueError("PusherStep can't be both, a jump and a pusher selection!")

        if (box_moved or moved_box_id) and is_jump:
            raise ValueError("PusherStep can't be both, a push and a jump!")

        if (box_moved or moved_box_id) and is_pusher_selection:
            raise ValueError("PusherStep can't be both, a push and a pusher selection!")

        self._box_moved: bool = False
        self._pusher_selected: bool = False
        self._pusher_jumped: bool = False
        self._pusher_id = pusher_id
        self.pusher_id = pusher_id
        self._moved_box_id: Optional[int] = None
        self.direction: Direction = direction
        self.is_current_pos = is_current_pos

        if box_moved:
            self.is_push_or_pull = True
        else:
            self.is_move = True

        if moved_box_id:
            self.moved_box_id = moved_box_id

        if is_jump:
            self.is_jump = is_jump

        if is_pusher_selection:
            self.is_pusher_selection = is_pusher_selection

    def __repr__(self):
        return "PusherStep({0}, box_moved={1})".format(
            str(self.direction), self.is_push_or_pull
        )

    def __str__(self):
        return (
            "PusherStep({0}, box_moved={1}, is_jump={2}, "
            "is_pusher_selection={3} pusher_id={4} box_id={5})"
        ).format(
            str(self.direction),
            self.is_push_or_pull,
            self.is_jump,
            self.is_pusher_selection,
            self.pusher_id,
            self.moved_box_id,
        )

    def __eq__(self, rv):
        return (
            self.direction == rv.direction
            and self.is_push_or_pull == rv.is_push_or_pull
            and self.is_pusher_selection == rv.is_pusher_selection
            and self.is_jump == rv.is_jump
        )

    def __ne__(self, rv):
        return not self == rv

    @property
    def moved_box_id(self) -> Optional[int]:
        """
        ID of box that was moved or None, depending on type of :class:`.PusherStep`
        """
        return self._moved_box_id if self.is_push_or_pull else None

    @moved_box_id.setter
    def moved_box_id(self, value: Optional[int]):
        """
        Updates ID of moved box and if this ID is valid, also changes this to
        push/pull. If removing ID, changes this to not-push/not-pull
        """
        if isinstance(value, int) and value >= Config.DEFAULT_PIECE_ID:
            self._moved_box_id = value
            self.is_push_or_pull = True
        else:
            self._moved_box_id = None
            self.is_push_or_pull = False

    @property
    def pusher_id(self) -> int:
        """ID of pusher that performed movement."""
        return self._pusher_id

    @pusher_id.setter
    def pusher_id(self, value: int):
        """
        Note:
            Pusher ID can't be None, thus we always set it to Config.DEFAULT_PIECE_ID
        """
        if isinstance(value, int) and value >= Config.DEFAULT_PIECE_ID:
            self._pusher_id = value
        else:
            self._pusher_id = Config.DEFAULT_PIECE_ID

    @property
    def is_move(self) -> bool:
        """
        True if pusher didn't move box, jump or changed focus to other pusher.
        """
        return (
            not self._box_moved
            and not self._pusher_selected
            and not self._pusher_jumped
        )

    @is_move.setter
    def is_move(self, value: bool):
        if value:
            self._box_moved = False
            self._pusher_jumped = False
            self._pusher_selected = False
            self._moved_box_id = None
        else:
            self._box_moved = True
            self._pusher_jumped = False
            self._pusher_selected = False

    @property
    def is_push_or_pull(self) -> bool:
        """True if pusher also moved a box."""
        return self._box_moved and not self._pusher_selected and not self._pusher_jumped

    @is_push_or_pull.setter
    def is_push_or_pull(self, value: bool):
        if value:
            self._box_moved = True
            self._pusher_jumped = False
            self._pusher_selected = False
        else:
            self._box_moved = False
            self._moved_box_id = None

    @property
    def is_pusher_selection(self) -> bool:
        """
        True if this move is part of change active pusher sequence in Multiban games.
        """
        return self._pusher_selected and not self._pusher_jumped and not self._box_moved

    @is_pusher_selection.setter
    def is_pusher_selection(self, value: bool):
        if value:
            self._pusher_selected = True
            self._box_moved = False
            self._pusher_jumped = False
            self._moved_box_id = None
        else:
            self._pusher_selected = False

    @property
    def is_jump(self) -> bool:
        """
        True if this move is part of pusher jump sequence in
        :attr:`.SolvingMode.REVERSE` games.
        """
        return self._pusher_jumped and not self._box_moved and not self._pusher_selected

    @is_jump.setter
    def is_jump(self, value: bool):
        if value:
            self._pusher_jumped = True
            self._pusher_selected = False
            self._box_moved = False
            self._moved_box_id = None
        else:
            self._pusher_jumped = False



def except_keys(dictionary, keys):
    if dictionary:
        key_set = set(dictionary.keys()) - set(keys)
        return {key: dictionary[key] for key in key_set}
    return {}


def inverted(dictionary):
    return dict(zip(dictionary.values(), dictionary.keys()))



from abc import ABC, abstractmethod


class MoverCommand(ABC):
    def __init__(self, mover):
        self._mover = mover
        self._moves = []
        self._rendered_moves = []

    @abstractmethod
    def redo(self):
        """Redoes movement command."""
        pass

    @abstractmethod
    def undo(self):
        """Undoes movement command."""
        pass

    @property
    def mover(self):
        return self._mover

    @property
    def rendered(self):
        """
        Sequence of :class:`.PusherStep` from last command execution (redo or undo).

        This sequence shows the moves like they would be rendered in GUI. Because of
        that, it shows different moves between undo and redo calls.

        See Also:
            `.Mover.last_move`
        """
        return self._rendered_moves

    @property
    def moves(self):
        """
        Sequence of :class:`.PusherStep` as a result of redo execution.

        This sequence shows the moves like they would be stored in :class:`.Snapshot`
        and remains same between undo and redo executions.
        """
        return self._moves


class SelectPusherCommand(MoverCommand):
    def __init__(self, mover, pusher_id):
        super().__init__(mover)
        self._new_pusher_id = pusher_id
        self._old_pusher_id = self._mover.selected_pusher

    def redo(self):
        self._mover.select_pusher(self._new_pusher_id)
        self._moves = self._mover.last_move
        self._rendered_moves = self._mover.last_move

    def undo(self):
        self._mover.select_pusher(self._old_pusher_id)
        self._rendered_moves = self._mover.last_move

    @property
    def old_pusher_id(self):
        return self._old_pusher_id

    @property
    def new_pusher_id(self):
        return self._new_pusher_id


class JumpCommand(MoverCommand):
    def __init__(self, mover, final_position):
        super().__init__(mover)
        self._initial_position = mover.board_manager.pusher_position(
            mover.selected_pusher
        )
        self._final_position = final_position

    def redo(self):
        self._mover.jump(self._final_position)
        self._moves = self._mover.last_move
        self._rendered_moves = self._mover.last_move

    def undo(self):
        self._mover.jump(self._initial_position)
        self._rendered_moves = self._mover.last_move

    @property
    def initial_position(self):
        return self._initial_position

    @property
    def final_position(self):
        return self._final_position


class MoveCommand(MoverCommand):
    def __init__(self, mover, direction):
        super().__init__(mover)
        self._direction = direction

    @property
    def direction(self):
        return self._direction

    def redo(self):
        self._mover.move(self._direction)
        self._moves = self._mover.last_move
        self._rendered_moves = self._mover.last_move

    def undo(self):
        self._mover.last_move = self._moves
        self._mover.undo_last_move()
        self._rendered_moves = self._mover.last_move



from __future__ import annotations

import io
import re
import textwrap
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Final, List, Optional, Pattern, Tuple, Union

import arrow

from .collection import Collection
from .hexoban import HexobanPuzzle, HexobanSnapshot
from .octoban import OctobanPuzzle, OctobanSnapshot
from .puzzle import Puzzle
from .snapshot import Snapshot
from .sokoban import SokobanPuzzle, SokobanSnapshot
from .trioban import TriobanPuzzle, TriobanSnapshot
from .utilities import is_blank

if TYPE_CHECKING:
    from ..game import Tessellation

_SELF_DIR = Path(__file__).absolute().resolve().parent
_SOK_FORMAT_SPEC_PATH = _SELF_DIR / "SOK_format_specification.txt"


class SOKFileFormat:
    @classmethod
    def read(
        cls,
        src: Union[io.StringIO, io.TextIOWrapper],
        dest: Collection,
        tessellation_hint: Optional[Tessellation] = None,
    ):
        from ..game import Tessellation

        SOKReader(src, dest, tessellation_hint or Tessellation.SOKOBAN).read()

    @classmethod
    def write(cls, src: Collection, dest: Union[io.StringIO, io.TextIOWrapper]):
        SOKWriter(dest).write(src)


class SOKReader:
    def __init__(
        self,
        src: Union[io.StringIO, io.TextIOWrapper],
        dest: Collection,
        tessellation_hint: Tessellation,
    ):
        self.src = src
        self.dest = dest
        self.supplied_tessellation_hint = tessellation_hint
        self._data: CollectionData

    def read(self):
        self._parse()
        self._copy()

    def _parse(self):
        self.src.seek(0, 0)
        self._data = CollectionData()
        self._split_input(self.src.readlines())
        self._parse_title_lines()
        self._parse_notes()

    def _copy(self):
        from ..game import Tessellation

        self.dest.title = self._data.title or ""
        self.dest.author = self._data.author or ""
        self.dest.created_at = self._data.created_at or ""
        self.dest.updated_at = self._data.updated_at or ""
        self.dest.notes = self._data.notes

        for puzzle_data in self._data.puzzles:
            if puzzle_data.tessellation == Tessellation.SOKOBAN:
                puzzle = SokobanPuzzle(board=puzzle_data.board)
                snapshot_klass = SokobanSnapshot
            elif puzzle_data.tessellation == Tessellation.HEXOBAN:
                puzzle = HexobanPuzzle(board=puzzle_data.board)
                snapshot_klass = HexobanSnapshot
            elif puzzle_data.tessellation == Tessellation.TRIOBAN:
                puzzle = TriobanPuzzle(board=puzzle_data.board)
                snapshot_klass = TriobanSnapshot
            elif puzzle_data.tessellation == Tessellation.OCTOBAN:
                puzzle = OctobanPuzzle(board=puzzle_data.board)
                snapshot_klass = OctobanSnapshot
            else:
                raise ValueError(
                    f"Missing implementation for {puzzle_data.tessellation}!"
                )

            for attr in {"title", "author", "boxorder", "goalorder", "notes"}:
                setattr(puzzle, attr, getattr(puzzle_data, attr))

            for snapshot_data in puzzle_data.snapshots:
                snapshot = snapshot_klass(moves_data=snapshot_data.moves_data or "")
                for attr in {"title", "solver", "notes"}:
                    setattr(snapshot, attr, getattr(snapshot_data, attr))
                puzzle.snapshots.append(snapshot)

            self.dest.puzzles.append(puzzle)

    def _split_input(self, input_lines: List[str]):
        first_board_line = first_index_of(input_lines, Puzzle.is_board)
        if first_board_line is not None:
            self._data.notes = input_lines[:first_board_line]
            remaining_lines = input_lines[first_board_line:]
        else:
            self._data.notes = input_lines
            remaining_lines = []

        self._split_puzzle_chunks(remaining_lines)
        self._split_snapshot_chunks()

    def _split_puzzle_chunks(self, lines: List[str]):
        remaining_lines = lines
        while remaining_lines:
            puzzle = PuzzleData()

            first_note_line = first_index_of(
                remaining_lines, lambda x: not Puzzle.is_board(x)
            )
            if first_note_line is not None:
                puzzle.board = "".join(remaining_lines[:first_note_line])
                remaining_lines = remaining_lines[first_note_line:]
            else:
                puzzle.board = "".join(remaining_lines)
                remaining_lines = []

            if len(remaining_lines) > 0:
                first_board_line = first_index_of(remaining_lines, Puzzle.is_board)

                if first_board_line is not None:
                    puzzle.notes = remaining_lines[:first_board_line]
                    remaining_lines = remaining_lines[first_board_line:]
                else:
                    puzzle.notes = remaining_lines
                    remaining_lines = []
            else:
                puzzle.notes = []

            self._data.puzzles.append(puzzle)

    def _split_snapshot_chunks(self):
        for puzzle in self._data.puzzles:
            remaining_lines = puzzle.notes

            first_moves_line = first_index_of(remaining_lines, Snapshot.is_snapshot)
            if first_moves_line is not None:
                puzzle.notes = remaining_lines[:first_moves_line]
                remaining_lines = remaining_lines[first_moves_line:]
            else:
                puzzle.notes = remaining_lines
                remaining_lines = []

            puzzle.snapshots = []

            while len(remaining_lines) > 0:
                snapshot = SnapshotData()

                first_note_line = first_index_of(
                    remaining_lines, lambda x: not Snapshot.is_snapshot(x)
                )
                if first_note_line is not None:
                    snapshot.moves_data = "".join(
                        moves_line.strip()
                        for moves_line in remaining_lines[:first_note_line]
                    )
                    remaining_lines = remaining_lines[first_note_line:]
                else:
                    snapshot.moves_data = "".join(
                        moves_line.strip() for moves_line in remaining_lines
                    )
                    remaining_lines = []

                if len(remaining_lines) > 0:
                    first_moves_line = first_index_of(
                        remaining_lines, Snapshot.is_snapshot
                    )

                    if first_moves_line is not None:
                        snapshot.notes = remaining_lines[:first_moves_line]
                        remaining_lines = remaining_lines[first_moves_line:]
                    else:
                        snapshot.notes = remaining_lines
                        remaining_lines = []
                else:
                    snapshot.notes = []

                puzzle.snapshots.append(snapshot)

    def _notes_before_puzzle(self, puzzle_index: int) -> List[str]:
        if puzzle_index == 0:
            return self._data.notes
        previous_puzzle = self._data.puzzles[puzzle_index - 1]
        if len(previous_puzzle.snapshots) > 0:
            return previous_puzzle.snapshots[-1].notes
        return previous_puzzle.notes

    def _notes_before_snapshot(
        self, puzzle_index: int, snapshot_index: int
    ) -> List[str]:
        puzzle = self._data.puzzles[puzzle_index]
        if snapshot_index == 0:
            return puzzle.notes
        return puzzle.snapshots[snapshot_index - 1].notes

    @staticmethod
    def _get_and_remove_title_line(notes: List[str]) -> str:
        """
        :: Titles                                                 ::
        ::   A title line is the last non-blank text line before  ::
        ::   a board, a saved game, or a solution, provided the   ::
        ::   line is preceded by a blank line or it is the only   ::
        ::   text line at this position in the file.              ::
        ::                                                        ::
        ::   Title lines are optional unless a single or a last   ::
        ::   text line from a preceding puzzle, saved game,       ::
        ::   solution, or file header can be mistaken for a title ::
        ::   line.                                                ::
        """
        candidate_index = last_index_of(notes, lambda x: not is_blank(x))
        if candidate_index is None:
            return ""

        preceding_index = None
        if candidate_index > 0:
            preceding_index = candidate_index - 1

        following_index = None
        if candidate_index < len(notes) - 1:
            following_index = candidate_index + 1

        preceding_ok = is_blank(notes[preceding_index]) if preceding_index else True

        following_ok = is_blank(notes[following_index]) if following_index else True

        if preceding_ok and following_ok:
            title_line = notes[candidate_index].strip()
            del notes[candidate_index]
            return title_line

        return ""

    def _parse_title_lines(self):
        for puzzle_index, puzzle in enumerate(self._data.puzzles):
            puzzle.title = self._get_and_remove_title_line(
                self._notes_before_puzzle(puzzle_index)
            )
            for snapshot_index, snapshot in enumerate(puzzle.snapshots):
                snapshot.title = self._get_and_remove_title_line(
                    self._notes_before_snapshot(puzzle_index, snapshot_index)
                )

    def _parse_notes(self):
        remaining_lines = SOKTags.extract_collection_attributes(
            self._data, self._data.notes
        )
        self._data.notes = self._cleanup_whitespace(remaining_lines)

        for puzzle_data in self._data.puzzles:
            remaining_lines = SOKTags.extract_puzzle_attributes(
                puzzle_data,
                puzzle_data.notes,
                self._data.header_tessellation_hint,
                self.supplied_tessellation_hint,
            )
            puzzle_data.notes = self._cleanup_whitespace(remaining_lines)

            for snapshot in puzzle_data.snapshots:
                remaining_lines = SOKTags.extract_snapshot_attributes(
                    snapshot, snapshot.notes
                )
                snapshot.notes = self._cleanup_whitespace(remaining_lines)

    @staticmethod
    def _cleanup_whitespace(lst) -> List[str]:
        i = first_index_of(lst, lambda x: not is_blank(x))
        if i is None:
            return []
        lst = lst[i:]

        i = last_index_of(lst, lambda x: not is_blank(x))
        if i is not None:
            lst = lst[: i + 1]

        return "\n".join(line.strip() for line in lst).split("\n")


@dataclass
class SnapshotData:
    moves_data: Optional[str] = None
    title: Optional[str] = None
    solver: Optional[str] = None
    notes: List[str] = field(default_factory=list)


@dataclass
class PuzzleData:
    board: Optional[str] = None
    tessellation: Optional[Tessellation] = None
    title: Optional[str] = None
    author: Optional[str] = None
    boxorder: Optional[str] = None
    goalorder: Optional[str] = None
    notes: List[str] = field(default_factory=list)
    snapshots: List[SnapshotData] = field(default_factory=list, init=False, repr=False)


@dataclass
class CollectionData:
    title: Optional[str] = None
    author: Optional[str] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None
    notes: List[str] = field(default_factory=list)
    header_tessellation_hint: Optional[Tessellation] = None
    puzzles: List[PuzzleData] = field(default_factory=list, init=False, repr=False)


class SOKTags:
    AUTHOR: Final[str] = "Author"
    TITLE: Final[str] = "Title"
    COLLECTION: Final[str] = "Collection"
    GOALORDER: Final[str] = "goalorder"
    BOXORDER: Final[str] = "boxorder"
    SOLVER: Final[str] = "Solver"
    VARIANT: Final[str] = "Game"
    DATE_CREATED: Final[str] = "Date Created"
    DATE_OF_LAST_CHANGE: Final[str] = "Date of Last Change"
    RAW_FILE_NOTES: Final[str] = "::"
    TAG_DELIMITERS: Final[str] = "=:"

    COLLECTION_TAGS: Final[Dict[str, str]] = {
        AUTHOR: "author",
        TITLE: "title",
        COLLECTION: "title",
        VARIANT: "",
        DATE_CREATED: "created_at",
        DATE_OF_LAST_CHANGE: "updated_at",
    }

    PUZZLE_TAGS: Final[Dict[str, str]] = {
        TITLE: "title",
        AUTHOR: "author",
        BOXORDER: "boxorder",
        GOALORDER: "goalorder",
        VARIANT: "",
    }

    SNAPSHOT_TAGS: Final[Dict[str, str]] = {
        TITLE: "title",
        AUTHOR: "solver",
        SOLVER: "solver",
    }

    @classmethod
    def extract_collection_attributes(
        cls, dest: CollectionData, notes: List[str]
    ) -> List[str]:
        from ..game import Tessellation

        remaining_lines = []
        tessellation = None
        for line in notes:
            if cls.is_raw_file_notes_line(line):
                continue

            was_tagged = False
            for tag, attr in cls.COLLECTION_TAGS.items():
                found, value = cls.get_tag_data(tag, line)
                if found:
                    if tag == cls.VARIANT:
                        tessellation = value
                    else:
                        setattr(dest, attr, value)
                    was_tagged = True
                    break

            if not was_tagged:
                remaining_lines.append(line)

        if not is_blank(tessellation):
            dest.header_tessellation_hint = Tessellation[tessellation.strip().upper()]

        return remaining_lines

    @classmethod
    def extract_puzzle_attributes(
        cls,
        dest: PuzzleData,
        notes: List[str],
        collection_header_tessellation_hint: Optional[Tessellation],
        supplied_tessellation_hint: Optional[Tessellation],
    ) -> List[str]:
        from ..game import Tessellation

        remaining_lines = []
        tessellation = None
        for line in notes:
            was_tagged = False
            for tag, attr in cls.PUZZLE_TAGS.items():
                found, value = cls.get_tag_data(tag, line)
                if found:
                    if tag == cls.VARIANT:
                        tessellation = value
                    else:
                        setattr(dest, attr, value)
                    was_tagged = True
                    break

            if not was_tagged:
                remaining_lines.append(line)

        if not is_blank(tessellation):
            dest.tessellation = Tessellation[tessellation.strip().upper()]
        elif collection_header_tessellation_hint is not None:
            dest.tessellation = collection_header_tessellation_hint
        elif supplied_tessellation_hint is not None:
            dest.tessellation = supplied_tessellation_hint
        else:
            dest.tessellation = Tessellation.SOKOBAN

        return remaining_lines

    @classmethod
    def extract_snapshot_attributes(
        cls, dest: SnapshotData, notes: List[str]
    ) -> List[str]:
        remaining_lines = []
        for line in notes:
            was_tagged = False
            for tag, attr in cls.SNAPSHOT_TAGS.items():
                found, value = cls.get_tag_data(tag, line)
                if found:
                    setattr(dest, attr, value)
                    was_tagged = True
                    break

            if not was_tagged:
                remaining_lines.append(line)

        return remaining_lines

    _TAG_SPLITTER: Final[Pattern] = re.compile(
        "|".join(map(re.escape, list(TAG_DELIMITERS)))
    )

    @classmethod
    def get_tag_data(cls, tag: str, line: str) -> Tuple[bool, Optional[str]]:
        """
        If it successfully extracts ``tag`` data from ``line``, it returns tuple
        ``(True, extracted_value). Otherwise, returns tuple ``(false, None)``.
        """
        components = cls._TAG_SPLITTER.split(str(line), 1)
        if components[0].strip().lower() == tag.strip().lower():
            return (True, "".join(components[1:]).strip())
        return (False, None)

    @classmethod
    def is_raw_file_notes_line(cls, line: str) -> bool:
        return line.lstrip().startswith(cls.RAW_FILE_NOTES)

    @classmethod
    def write_tagged(
        cls, dest: Union[io.StringIO, io.TextIOWrapper], tag: str, data: Optional[str]
    ) -> bool:
        if not data or is_blank(data) or is_blank(tag):
            return False
        dest.write(tag.strip() + ": " + data.rstrip() + "\n")
        return True


class SOKWriter:
    def __init__(self, dest: Union[io.StringIO, io.TextIOWrapper]):
        self.dest = dest

    def write(self, src: Collection):
        self._write_collection_header(src)
        for puzzle in src.puzzles:
            self._write_puzzle(puzzle)

    def _write_collection_header(self, src: Collection):
        for line in open(_SOK_FORMAT_SPEC_PATH):
            self.dest.write(line.rstrip() + "\n")

        SOKTags.write_tagged(
            self.dest,
            SOKTags.DATE_CREATED,
            src.created_at.strip() or arrow.utcnow().isoformat(),
        )
        SOKTags.write_tagged(
            self.dest,
            SOKTags.DATE_OF_LAST_CHANGE,
            src.updated_at.strip() or arrow.utcnow().isoformat(),
        )

        self.dest.write(
            "::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::" + "\n\n"
        )

        written = False
        written = (
            SOKTags.write_tagged(self.dest, SOKTags.COLLECTION, src.title) or written
        )
        written = SOKTags.write_tagged(self.dest, SOKTags.AUTHOR, src.author) or written

        puzzle_collection_notes = "\n".join(src.notes)
        if not is_blank(puzzle_collection_notes):
            self.dest.write(puzzle_collection_notes.rstrip() + "\n")
            written = True

        if written:
            self.dest.write("\n")

    def _write_puzzle(self, src: Puzzle):
        from ..game import Tessellation

        if is_blank(src.board):
            return

        if not is_blank(src.title):
            self.dest.write(src.title.strip() + "\n\n")

        self.dest.write(src.board.rstrip() + "\n\n")

        written = False

        if src.tessellation != Tessellation.SOKOBAN:
            written = (
                SOKTags.write_tagged(
                    self.dest, SOKTags.VARIANT, str(src.tessellation.value).lower()
                )
                or written
            )

        if not is_blank(src.boxorder) and not is_blank(src.goalorder):
            written = (
                SOKTags.write_tagged(self.dest, SOKTags.BOXORDER, src.boxorder)
                or written
            )
            written = (
                SOKTags.write_tagged(self.dest, SOKTags.GOALORDER, src.goalorder)
                or written
            )

        if not is_blank(src.author):
            written = (
                SOKTags.write_tagged(self.dest, SOKTags.AUTHOR, src.author) or written
            )

        puzzle_notes = "\n".join(src.notes)
        if not is_blank(puzzle_notes):
            self.dest.write(puzzle_notes.rstrip() + "\n")
            written = True

        if written:
            self.dest.write("\n")

        for snapshot in src.snapshots:
            self._write_snapshot(snapshot)

    def _write_snapshot(self, src: Snapshot):
        if is_blank(src.moves_data):
            return

        if not is_blank(src.title):
            self.dest.write(src.title.strip() + "\n")

        self.dest.write("\n".join(textwrap.wrap(src.moves_data.strip(), 70)) + "\n")

        written = SOKTags.write_tagged(self.dest, SOKTags.SOLVER, src.solver)

        snap_notes = "\n".join(src.notes)
        if not is_blank(snap_notes):
            self.dest.write(snap_notes.rstrip() + "\n")
            written = True

        if written:
            self.dest.write("\n")


def first_index_of(lst, predicate):
    return next((index for index, elem in enumerate(lst) if predicate(elem)), None)


def last_index_of(lst, predicate):
    candidate_index = first_index_of(reversed(lst), predicate)
    if candidate_index is not None:
        return len(lst) - candidate_index - 1
    return None
